#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
from scipy.stats import sem
import seaborn as sns
import matplotlib.pyplot as plt

class Environment(object):

    def __init__(self, cards=None, nblocks=2):
        self.cards = cards
        self.nblocks = nblocks
        self.set_environment()

    def set_environment(self):
        self.trials = np.array([self.cards.index.values]*self.nblocks).flatten()
        self.ntrials = len(self.trials)
        self.nalt = len(self.cards.iloc[1:].columns)
        self.names = np.sort(self.cards.columns.values)


class Agent(Environment):

    def __init__(self, ap=.1, an=.1, b=5, cards=None, niter=10, nblocks=2):

        self.updateQ = lambda q, winner, r, A: q[winner][-1] + A*(r - q[winner][-1])
        self.updateP = lambda q, name, b: np.exp(b*q[name][-1])/np.sum([np.exp(b*q[k][-1]) for k in q.keys()])

        super(Agent, self).__init__(cards=cards, nblocks=nblocks)

        if any(np.size(v)>1 for v in [ap, an, b]):
            self.track_params(niter=niter, ap=ap, an=an, b=b)
        else:
            self.set_params(ap=ap, an=an, b=b)


    def set_params(self, ap, an, b):
        self.ap = ap
        self.an = an
        self.b = b
        self.choices = []


    def track_params(self, niter=10, ap=.1, an=.1, b=5):

        blks = np.arange(niter)
        param_names = ['ap', 'an', 'b']
        param_values = [ap, an, b]

        for i, pvalue in enumerate(param_values):
            if not hasattr(pvalue, '__iter__'):
                param_values[i] = [pvalue]

        apos, aneg, beta = param_values
        perm_param_values = list(itertools.product(apos, aneg, beta))

        nsets = len(perm_param_values)
        blocksdf = pd.DataFrame(perm_param_values, columns=param_names, index=np.arange(nsets))

        self.blocksdf = pd.concat([blocksdf]*niter)
        self.blocksdf.reset_index(inplace=True)
        self.blocksdf.rename(columns={'index':'block'}, inplace=True)
        self.blocksdf['P']=0
        self.blocksdf['Q']=0


    def iter_params(self):

        for i in self.blocksdf.index.values:
            ap, an, b = self.blocksdf.loc[i, ['ap', 'an', 'b']].values
            self.set_params(ap=ap, an=an, b=b)

            P, Q = self.simulate_task(return_scores=True)
            self.blocksdf.loc[i, 'P'] = P
            self.blocksdf.loc[i, 'Q'] = Q


    def simulate_task(self, return_scores=False):

        self.qdict={k:[0] for k in self.names}
        self.choice_prob={k:[1./self.nalt] for k in self.names}

        for t in self.trials:
            rew_vals = self.cards.iloc[t, :].values
            qvals = np.array([self.qdict[name][-1] for name in self.names])
            pvals = np.array([self.choice_prob[name][-1] for name in self.names])
            winner = np.random.choice(np.arange(self.nalt), p=pvals)
            wname = self.names[winner]

            r = rew_vals[winner]
            q = qvals[winner]
            rpe = r - q

            if rpe>0:
                alpha=self.ap
            else:
                alpha=self.an

            Qup = q + (alpha * rpe)
            self.qdict[wname].append(Qup)
            self.choice_prob[wname].append(self.updateP(self.qdict, wname, self.b))

            for loser in self.names[self.names!=wname]:
                self.qdict[loser].append(self.qdict[loser][-1])
                self.choice_prob[loser].append(self.updateP(self.qdict, loser, self.b))
            self.choices.append(winner)

        if return_scores:
            return self.igt_scores()

    def igt_scores(self):

        ch = np.asarray(self.choices)
        A = ch[ch==0].size
        B = ch[ch==1].size
        C = ch[ch==2].size
        D = ch[ch==3].size

        # payoff (P) score
        P = (C+D) - (A+B)
        # sensitivity (Q) score
        Q = (B+D) - (A+C)

        return [P, Q]

    def plot_summary(self, titles=['Order of Choices','Number of Choices per Card', 'Change in Q(card)',
        'Change in P(card)']):
        sns.set(style='darkgrid', context='paper', font_scale=1.4)
        f, axes = plt.subplots(2, 2, figsize=(10,8))
        a1, a2, a3, a4 = axes.flatten()

        a1.plot(self.choices, lw=0, marker='o')
        a1.set_ylim(-.5, 3.5); a1.set_yticks(np.arange(self.nalt))
        a1.set_yticklabels(self.names.upper())

        a2.hist(np.asarray(self.choices))
        a2.set_xticks(np.arange(self.nalt))
        a2.set_xticklabels(self.names.upper())

        for n in idx.names:
            a3.plot(self.qdict[n], label=n)
            a4.plot(self.choice_prob[n], label=n)

        a3.legend()
        a4.legend()

        for i, ax in enumerate(axes.flatten()):
            ax.set_title(titles[i])

        f.subplots_adjust(hspace=.35)



#!usr/bin/env python
from __future__ import division
import pandas as pd
import numpy as np
from numpy import array

class DataHandler(object):

    def __init__(self, model):

        self.model = model
        self.data = model.data
        self.inits = model.inits
        self.idx = model.idx
        self.nidx = model.nidx

        self.kind = model.kind
        self.fit_on = model.fit_on
        self.percentiles = model.percentiles
        self.groups = model.groups

        self.conds = model.conds
        self.nconds = model.nconds
        self.levels = model.levels
        self.nlevels = model.nlevels

        self.nrows = self.nidx * self.nlevels * self.nconds
        self.grpData = self.data.groupby(self.groups)


    def make_dataframes(self):
        """ Generates the following dataframes and arrays:
        ::Arguments::
              qp_cols:
                    header for observed/fits dataframes
        ::Returns::
              None (All dataframes and vectors are stored in dict and assigned
              as <dframes> attr)

        observed (list of ndarrays):
            list containing ndarrays (ncond x n_observed_data_points) entered into thec
            costfx during fitting. If self.fit_on=='average', self.observed will
            contain a single ndarray (ncond x average_data). If self.fit_on=='subjects',
            self.observed will contain 1 ndarray for each subject and fits will be performed
            iteratively
        observedDF (DF):
              Contains Prob and RT quant. for each subject
              used to calc. cost fx weights
        fits (DF):
              empty DF shaped like observed DF, used to store simulated
              predictions of the optimized model
        fitinfo (DF):
              stores all opt. parameter values and model fit statistics
        """

        self.make_observed_groupDFs()

        if self.fit_on=='subjects':
            idxdf = lambda idx: self.observedDF[self.observedDF['idx']==idx]
            observed = [idxdf(idx).dropna(axis=1).loc[:, 'acc':].values for idx in self.idx]
            observedflat = [idxdf(idx).dropna(axis=1).mean().loc['acc':].values for idx in self.idx]

        elif self.fit_on=='average':
            observed = [self.observedDF.groupby(self.conds).mean().loc[:, 'acc':].values]
            observedflat = [self.observedDF.mean().loc['acc':].values]

        # elif self.fit_on == 'bootstrap':
        #     observed = self.idxData.apply(self.resample_data)

        # Get rid of any extra dimensions
        self.observed = [obs_i.squeeze() for obs_i in observed]
        self.observed_flat = [obsF_i.squeeze() for obsF_i in observedflat]


    def make_observed_groupDFs(self):
        """ concatenate all idx data vectors into a dataframe
        """

        idxdf_cols, obsdf_cols, infodf_cols = self.__get_headers__()

        data = self.data
        ncols = len(obsdf_cols)
        nrows = self.nrows
        index = range(nrows)

        self.datdf = self.grpData.apply(self.model.rangl_data).sortlevel(0)
        self.dfvals = [self.datdf.values[i].astype(float) for i in index]

        self.observedDF = pd.DataFrame(np.zeros((nrows, ncols))*np.nan, columns=obsdf_cols, index=index)
        self.observedDF.loc[:, self.groups] = self.datdf.reset_index()[self.groups].values
        self.fits = self.observedDF.copy()

        for rowi in range(nrows):
            self.observedDF.loc[rowi, idxdf_cols[rowi]] = self.dfvals[rowi]

        # GENERATE DF FOR FIT RESULTS
        self.fitinfo = pd.DataFrame(columns=infodf_cols, index=index)
        self.fitinfo.loc[:, self.groups] = self.observedDF.loc[:, self.groups]


    def __get_headers__(self):

        obsdf_cols = np.hstack([self.groups + ['acc']]).tolist()
        cq = ['c' + str(int(n * 100)) for n in self.percentiles]
        eq = ['e' + str(int(n * 100)) for n in self.percentiles]
        qcols = cq + eq

        if hasattr(self.model, 'ssd'):
            ssd_list = [np.asarray(issd*1000, dtype=np.int) for issd in self.model.ssd]
            idxdf_cols = [['acc'] + issd.tolist() + qcols for issd in ssd_list]
            all_ssds = np.unique(np.hstack(ssd_list))
            obsdf_cols = obsdf_cols + all_ssds.tolist()
        else:
            idxdf_cols = [np.hstack(['acc'], qcols).tolist()]*nrows

        obsdf_cols = obsdf_cols + qcols
        params = np.sort(list(self.inits))
        fit_cols = ['nfev', 'nvary', 'df', 'chi', 'rchi', 'logp', 'AIC', 'BIC', 'cnvrg']
        fitdf_cols = np.hstack([self.groups, params, fit_cols]).tolist()

        return idxdf_cols, obsdf_cols, fitdf_cols


    def finfo_to_params(self, finfo='./finfo.csv', pc_map=None):

        pnames = set(['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso'])
        if isinstance(finfo, str):
            finfo = pd.read_csv(finfo, header=None, names=['id', 'vals'], index_col=0)
            finfo = pd.Series(finfo.to_dict()['vals'])
        elif isinstance(finfo, dict):
            finfo = pd.Series(finfo)
        elif isinstance(finfo, pd.Series):
            pass

        plist = list(pnames.intersection(finfo.index))
        params = {k: finfo[k] for k in plist}

        if pc_map != None:
            for pkey, pclist in pc_map.items():
                params[pkey] = array([params[pkc] for pkc in pclist])
        return params


    def rwr(self, X, get_index=False, n=None):
        """
        Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
        """
        if isinstance(X, pd.Series):
            X = X.copy()
            X.index = range(len(X.index))
        if n == None:
            n = len(X)
        resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
        X_resample = (X[resample_i])
        if get_index:
            return resample_i
        else:
            return X_resample


    def resample_data(self, n=120, groups=['ssd']):
        """ generates n resampled datasets using rwr()
        for bootstrapping model fits
        """

        df = self.data.copy()
        tb = self.tb

        bootlist = list()
        if n == None:
            n = len(df)

        for level, level_df in df.groupby(groups):
            boots = level_df.reset_index(drop=True)
            orig_ix = np.asarray(boots.index[:])
            resampled_ix = self.rwr(orig_ix, get_index=True, n=n)
            bootdf = level_df.irow(resampled_ix)
            bootlist.append(bootdf)

        # concatenate and return all resampled conditions
        return self.model.rangl_data(pd.concat(bootlist))


    def extract_popt_fitinfo(self, finfo=None):
        """ takes optimized dict or DF of vectorized parameters and
        returns dict with only depends_on.keys() containing vectorized vals.
        Is accessed by fit.Optimizer objects after optimization routine.

        ::Arguments::
              finfo (dict/DF):
                    finfo is dict if self.fit_on is 'average'
                    and DF if self.fit_on is 'subjects' or 'bootstrap'
                    contains optimized parameters
        ::Returns::
              popt (dict):
                    dict with only depends_on.keys() containing
                    vectorized vals
        """
        if finfo is None:
            try:
                finfo = self.model.fitinfo.mean()
            except Exception:
                finfo = self.model.fitinfo

        finfo = dict(deepcopy(finfo))
        popt = dict(deepcopy(self.inits))
        pc_map = self.model.pc_map

        for pkey in list(popt):
            if pkey in list(self.depends_on):
                popt[pkey] = np.array([finfo[pc] for pc in pc_map[pkey]])
                continue
            popt[pkey] = finfo[pkey]

        return popt



#!/usr/local/bin/env python
from __future__ import division
from future.utils import listvalues
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from radd.tools import messages, theta
from radd.tools.analyze import assess_fit
from radd.models import Simulator
from lmfit import minimize, fit_report
from radd.CORE import RADDCore
from scipy.optimize import basinhopping
from numpy.random import uniform


class BasinBounds(object):
    """ sets conditions for step acceptance during
    basinhopping optimization routine
    Arguments:
        xmin (list): lower boundaries for each parameter
        xmax (list): upper boundaries for each parameter
    """
    def __init__(self, xmin, xmax):
        self.xmin = np.array(xmin)
        self.xmax = np.array(xmax)

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmin = bool(np.all(x >= self.xmin))
        tmax = bool(np.all(x <= self.xmax))
        return tmin and tmax


class HopStep(object):
    """ scale stepsize of basinhopping optimization according to
    individual parameters ranges (smaller steps for more sensitive params).
    See theta.get_stepsize_scalars() for parameter <--> scalar mapping.

    Arguments:
        keys (list): list of parameter names
        stepsize (list): initial stepsize
    """
    def __init__(self, keys, nlevels=1, stepsize=0.05):
        self.stepsize_scalars = theta.get_stepsize_scalars(keys, nlevels)
        self.stepsize = stepsize
        self.np = len(self.stepsize_scalars)

    def __call__(self, x):
        s = self.stepsize
        ss = self.stepsize_scalars
        x = np.array([x[i] + uniform(-ss[i]*s, ss[i]*s) for i in range(self.np)])
        return x


class Optimizer(object):
    """ Optimizer class acts as interface between Model and Simulator objects.
    Structures fitting routines so that Models are first optimized with the full set of
    parameters free, data collapsing across conditions.

    The fitted parameters are then used as the initial parameters for fitting conditional
    models with only a subset of parameters are left free to vary across levels of a given
    experimental conditions.

    Parameter dependencies are specified when initializing Model object via
    <depends_on> arg (i.e.{parameter: condition})

    Handles fitting routines for models of average, individual subject, and bootstrapped data
    """

    def __init__(self, fitparams=None, basinparams=None, inits=None, kind='xdpm', depends_on=None, pc_map=None):

        self.inits=inits
        self.fitparams = fitparams
        self.basinparams = basinparams
        self.kind = kind

        self.pc_map = pc_map
        self.pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso']
        self.pvc = deepcopy(['a', 'tr', 'v', 'xb'])


    def get_best_inits(self, pkeys=None, nbest=10, nrand_samples=500):
        """ test a large sample of random parameter values
        and submit <nbest> to hop_around() global optimization
        """
        if pkeys is None:
            pkeys = np.sort(list(self.inits))

        rinits = theta.random_inits(pkeys, ninits=nrand_samples, kind=self.kind)
        all_inits = [{pk: rinits[pk][i] for pk in pkeys} for i in range(nrand_samples)]

        fmin_all = [self.simulator.__cost_fx__(inits_i) for inits_i in all_inits]
        fmin_series = pd.Series(fmin_all)

        best_inits_index = fmin_series.sort_values().index[:nbest]
        best_inits = [all_inits[i] for i in best_inits_index]
        return best_inits


    def hop_around(self, p):
        """ initialize model with niter randomly generated parameter sets
        and perform global minimization using basinhopping algorithm
        ::Arguments::
            p (dict):
                parameter dictionary
        ::Returns::
            parameter set with the best fit
        """

        bp = self.basinparams
        pkeys = np.sort(list(p))

        # get cost fmin for default inits
        p0 = dict(deepcopy(p))
        p_init = dict(deepcopy(p))
        fmin0 = self.simulator.__cost_fx__(p_init)
        print("default inits: fmin0=%.9f" % (fmin0))

        # sample random inits and select best of
        best_inits = self.get_best_inits(pkeys, nbest=bp['nrand_inits'], nrand_samples=bp['nrand_samples'])

        xpopt, xfmin = [], []
        for params in best_inits:
            popt, fmin = self.run_basinhopping(p=params, is_flat=True)
            xpopt.append(popt)
            xfmin.append(fmin)

        # get the global basin and
        # corresponding parameter estimates
        ix_min = np.argmin(xfmin)
        popt_best = xpopt[ix_min]
        fmin_best = xfmin[ix_min]

        # compare global basin (fmin_best) to
        # fmin using default inits (fmin0)
        if fmin_best >= fmin0:
            basin_decision = "USING DEFAULT INITS: fmin_inits=%.9f, next_best=%.9f" % (fmin0, fmin_best)
            print(basin_decision)
            return p0
        else:
            basin_decision = "NEW GLOBAL MINIMUM: fmin_new=%.9f; fmin_inits=%9f)" % (fmin_best, fmin0)
            print(basin_decision)
            return popt_best


    def run_basinhopping(self, p, is_flat=True):
        """ uses fmin_tnc in combination with basinhopping to perform bounded global
         minimization of multivariate model
        ::Arguments::
            is_flat (bool <True>):
              if True, optimize all params in p
        """
        bp = self.basinparams
        fp = self.fitparams

        if is_flat:
            basin_keys = np.sort(list(p))
            xp = dict(deepcopy(p))
            basin_params = theta.all_params_to_scalar(xp)
            nl = 1
        else:
            basin_keys = np.sort(list(self.pc_map))
            basin_params = deepcopy(p)
            nl = fp['y'].ndim

        self.simulator.__prep_global__(basin_params=basin_params, basin_keys=basin_keys)
        # make list of init values for all pkeys included in fit
        x0 = np.hstack(np.hstack([basin_params[pk]*np.ones(nl) for pk in basin_keys]))

        # define parameter boundaries for all params in pc_map.keys()
        # to be used by basinhopping minimizer & tnc local optimizer
        xmin, xmax = theta.format_basinhopping_bounds(basin_keys, nlevels=nl)
        tncbounds = theta.format_local_bounds(xmin, xmax)
        mkwargs = {"method": bp['method'], 'bounds': tncbounds, 'tol': bp['tol'],
            'options': {'xtol': bp['tol'], 'ftol': bp['tol'], 'maxiter': bp['maxiter']}}

        # define custom take_step and accept_test functions
        accept_step = BasinBounds(xmin, xmax)
        custom_step = HopStep(basin_keys, nlevels=nl, stepsize=bp['stepsize'])

        # run basinhopping on simulator.basinhopping_minimizer func
        out = basinhopping(self.simulator.basinhopping_minimizer, x0=x0, minimizer_kwargs=mkwargs, take_step=custom_step, accept_test=accept_step, stepsize=bp['stepsize'], niter=bp['niter'], niter_success=bp['niter_success'], interval=bp['interval'], disp=bp['disp'])

        xopt = out.x
        funcmin = out.fun
        if nl > 1:
            xopt = [array([xopt]).reshape(len(basin_keys), nl).squeeze()]
        for i, k in enumerate(basin_keys):
            p[k] = xopt[i]
        return p, funcmin


    def gradient_descent(self, inits=None, is_flat=True):
        """ Optimizes parameters following specified parameter
        dependencies on task conditions (i.e. depends_on={param: cond})
        """

        fp = self.fitparams
        if inits is None:
            inits = dict(deepcopy(self.inits))

        opt_kws = {'disp': fp['disp'], 'xtol': fp['tol'], 'ftol': fp['tol'],
        'maxiter': fp['maxiter'], 'maxfev': fp['maxfev']}

        # GEN PARAMS OBJ & OPTIMIZE THETA
        lmParams = theta.loadParameters(inits=inits, pc_map=self.pc_map, is_flat=is_flat, kind=self.kind)
        optmod = minimize(self.simulator.__cost_fx__, lmParams, method=fp['method'], options=opt_kws)

        # gen dict of opt. params
        finfo = dict(deepcopy(optmod.params.valuesdict()))
        popt = dict(deepcopy(finfo))
        y = self.simulator.y
        wts = self.simulator.wts
        yhat_iter = [self.simulator.sim_fx(popt) for i in range(100)]
        yhat = np.mean(yhat_iter, axis=0)

        # ASSUMING THE weighted SSE is CALC BY COSTFX
        finfo['chi'] = np.sum(wts * (y.flatten() - yhat)**2)
        finfo['ndata'] = len(yhat)
        finfo['cnvrg'] = optmod.success
        finfo['nfev'] = optmod.nfev
        finfo['nvary'] = len(optmod.var_names)
        finfo = assess_fit(finfo)
        print(finfo['cnvrg'])

        popt = theta.all_params_to_scalar(popt, exclude=list(self.pc_map))
        log_arrays = {'y': y, 'yhat': yhat, 'wts': wts}
        param_report = fit_report(optmod.params)
        messages.logger(param_report=param_report, finfo=finfo, pdict=popt, depends_on=fp['depends_on'], log_arrays=log_arrays, is_flat=is_flat, kind=self.kind, fit_on=fp['fit_on'], dynamic=fp['dynamic'], pc_map=self.pc_map)

        return yhat, finfo, popt



#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
from collections import OrderedDict
import numpy as np
from numpy import array
from numpy import newaxis as na
from numpy.random import randint
from numpy import concatenate as concat
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from radd import build
from radd.tools import vis
from radd.tools.colors import get_cpals
from radd.tools.theta import get_xbias_theta
from radd.tools.messages import describe_model
from radd.models import Simulator

class BOLD(Simulator):
    """ Simulated BOLD response // neural activity
    from optimized RADD models

    Largley based on mapping lists of arrays from single
    conditions to lambda/def functions, due to many non-uniform
    ndarray transformations (i.e., boolean indexing that leads to
    different lengths along a given axis of the full DVg)

    """

    def __init__(self, model, sim_xbias=False):

        if hasattr(model, 'popt'):
            self.p = model.popt
            try:
                for pkey, pkc_list in model.pc_map.items():
                    self.p[pkey] = np.array([model.popt[pkc] for pkc in pkc_list])
            except KeyError:
                pass
        elif sim_xbias:
            self.p = get_xbias_theta(model)
        else:
            model.fit_whole_model = False
            try:
                # model.__get_default_inits__()
                self.p = model.__get_optimized_params___()
            except Exception:
                self.p = model.inits

        # GENERATE MODEL SIMULATOR
        super(BOLD, self).__init__(model=model, inits=self.p, pc_map=model.pc_map, kind=model.kind, dt=.001)

        self.ncond = model.ncond
        self.depends_on = model.depends_on
        self.simulate = self.sim_fx
        self.labels = model.labels
        self.__init_process_functions__()

    def __init_process_functions__(self):

        #############################################################
        # LAMBDA FUNC: GET_RT FOR ALL BOUNDARY CROSSES, 999 OTHERWISE
        # x[0] = DVg (go accumulator: all trials in single condition)
        # x[1] = a   (boundary, single condition)
        # x[2] = tr  (onset time, cond i)
        self.get_rt = lambda x: np.where(x[0].max(axis=1) >= x[1], x[2] + np.argmax(x[0] >= x[1], axis=1) * self.dt, 999)
        self.get_ssrt = lambda dvs, delay: np.where(dvs.min(axis=3) <= 0, delay + np.argmax(dvs <= 0, axis=3) * self.dt, 999)

        ###############################################################
        # LAMBDA FUNC: GET_GO_TRACES FOR ALL THAT CROSS BOUND BEFORE TB
        # x[0] = DVg (go accumulator: all trials in single condition)
        # x[1] = RT  (all trials in single condition)
        # x[1] = timeboundary (simulator.tb)
        self.get_go_traces = lambda x: x[0][np.where(x[1] < x[2], True, False)]
        self.get_ng_traces = lambda x: x[0][
            np.where(x[1] >= x[2], True, False)]

        #######################################################################
        # LAMBDA FUNC: CAP_ROLL_TRACES cap traces at bound and roll nans to back (opt. decay)
        # x[0] = DVg (go accumulator: all trials in single condition)
        # x[1] = a (boundary, single condition)
        # x[2] = Tg (ntime points, single condition)
        # x[3] = BOOL (decay if True, <default False>)
        self.cap_roll_traces = lambda x: self.drop_n_roll(self.cap_n_bound(x[0], x[1], x[2], x[3]))

        #######################################################################
        # LAMBDA FUNC: GET_HEMO get last occurence of numeric vals along cumsum axis=1
        # df = BOLD DF Super (all trials cumsum(axis=1), single condition)
        #self.get_hemo = lambda df: df.cumsum().apply(self.__ix_lastnum__, axis=1)
        #self.get_first_numeric = lambda df: df.apply(self.__ix_firstnum__, axis=1)
        def get_last_numeric(x):
            if x.last_valid_index() is None:
                return None
            else:
                return x[x.last_valid_index()]

        self.get_hemo = lambda d: d.cumsum(axis=0).apply(get_last_numeric, axis=0).dropna().values

    def generate_dpm_traces(self):
        """ Get go,ssrt using same function as proactive then use the indices for
        correct go and correct stop trials to slice summed, but momentary, dvg/dvs evidence vectors
        (not accumulated over time just a basic sum of the vectors)
        """
        # ensure parameters are all vectorized
        self.p = self.vectorize_params(self.p)
        Pg, Tg = self.__update_go_process__(self.p)
        Ps, Ts = self.__update_stop_process__(self.p)

        self.bound = self.p['a']
        self.onset = self.p['tr']

        ncond = self.ncond
        ntot = self.ntot
        dx = self.dx
        base = self.base
        self.nss_all = self.nss
        nssd = self.nssd
        ssd = self.ssd
        nss = self.nss_all / self.nssd
        xtb = self.xtb

        get_ssbase = lambda Ts, Tg, DVg: array([[DVc[:nss / nssd, ix] for ix in np.where(Ts < Tg[i], Tg[i] - Ts, 0)] for i, DVc in enumerate(DVg)])[:, :, :, na]

        self.gomoments = xtb[:, na] * np.where((rs((ncond, ntot, Tg.max())).T < Pg), dx, -dx).T
        self.ssmoments = np.where(rs((ncond, nssd, nss, Ts.max())) < Ps, dx, -dx)
        DVg = base[:, na] + xtb[:, na] * np.cumsum(self.gomoments, axis=2)
        self.dvg = DVg[:, :nss_all, :]
        self.dvs = self.get_ssbase(Ts, Tg, DVg) + np.cumsum(self.ssmoments, axis=3)

        dg = self.gomoments[:, nss_all:, :].reshape(ncond, nssd, nss, Tg.max())
        ds = self.ssmoments.copy()

        ss_list, go_list = [], []
        for i, (s, g) in enumerate(zip(ds, dg)):
            diff = Ts - Tg[i]
            pad_go = diff[diff > 0]
            pad_ss = abs(diff[diff < 0])
            go_list.extend([concat((np.zeros((nss, gp)), g[i]), axis=1) for gp in pad_go])
            go_list.extend([g[x] for x in range(len(pad_ss))])

            ss_list.extend([concat((np.zeros((nss, abs(spad))), s[i, :, -(Tg[i] - spad):]), axis=1) for spad in pad_ss])
            ss_list.extend([s[i, :, -tss:] for tss in Ts[:len(pad_go)]])

        r = map((lambda x: np.cumsum((x[0] + x[1]), axis=1)), zip(go_list, ss_list))

    def generate_pro_traces(self, ntrials=500):
        """ ensures parameters are vectorized and sets
        bound and onset attr. before generating simulated traces
        """
        # ensure parameters are all vectorized
        self.p = self.vectorize_params(self.p)

        # init timebound, bound, onset-time attr
        self.bound = self.p['a']
        self.onset = self.p['tr']
        self.ntot = ntrials  # *self.ncond
        # simulate decision traces
        self.dvg = self.sim_fx(self.p, analyze=False)

    def simulate_bold(self, hemodynamic=True, ntrials=500, decay=False, get_dfs=False, shape='long', savestr='./',  save=False):
        """ gets RT of boundary crossing for boolean selecting
        go traces which are then filtered and stored in DF

        Traces (trials) that meet the criteria:
              :: [DVg(trial_vals) >= a] & [DVg(trial_rt)<tb] ::
        are then
              ::capped:: setting all values after initial threshold to NAN
              ::rolled:: shifted so all NAN values are moved to end of trace

        ::Arguments::
              hemodynamic (bool <True>):
                    simulate superposition property (cumumulative accumulator)
              decay (bool <False>):
                    concatenate mirror image of input array
                    along axis 1 (i.e., bold up, decay down)
              get_dfs (bool <False>):
                    return list of simulated bold dataframes
        """
        if decay:
            self.decay = "wdecay"
        else:
            self.decay = "nodecay"
        if 'pro' in self.kind:
            # simulate decision traces
            self.generate_pro_traces(ntrials=ntrials)
            self.Tg = np.ceil((self.tb - self.onset) / self.dt).astype(int)
        else:
            # simulate decision traces
            self.generate_dpm_traces()

        # zip dvg[ncond, ntrials, ntime], bound[a_c..a_ncond],
        # onset[tr_c..tr_ncond]
        zipped_input_rt = zip(self.dvg, self.bound[:, na], self.onset[:, na])
        # get rt for all conditions separately where boundary crossed [DVg >=
        # a]
        rt = np.asarray(map(self.get_rt, zipped_input_rt))

        # zip dvg[ncond, ntrials, ntime], rt[rt_c..rt_ncond], [tb]*ncond
        zipped_rt_traces = zip(self.dvg, rt, [self.tb] * len(rt))
        # boolean index traces so that DVg[RT<=TB]
        go_trial_arrays = map(self.get_go_traces, zipped_rt_traces)

        if 'pro' in self.kind:
            # zip dvg(resp=1), bound[a_c..a_ncond], [decay]*ncond
            ng_trial_arrays = map(self.get_ng_traces, zipped_rt_traces)
        elif 'dpm' in self.kind:
            ssrt = self.get_ssrt(self.dvs, self.ssd[:, None])
            strial_gorts = rt[:, :self.nss_all].reshape(self.ncond, self.nssd, self.nss)
            # zip DVs, SSRT, ssGoRT
            zipped_rt_traces = zip(self.dvs, ssrt, strial_gorts)
            ng_trial_arrays = map(self.get_ng_traces, zipped_rt_traces)

        zipped_go_caproll = zip(go_trial_arrays, self.bound, self.Tg, [decay] * self.ncond)
        zipped_ng_caproll = zip(ng_trial_arrays, self.bound, self.Tg, [decay] * self.ncond)
        # generate dataframe of capped and time-thresholded go traces
        self.go_traces = map(self.cap_roll_traces, zipped_go_caproll)
        self.ng_traces = map(self.cap_roll_traces, zipped_ng_caproll)

        if hemodynamic:
            self.make_bold_dfs(shape=shape, save=save)
            self.mean_go_traces = [gt.mean(axis=1).dropna().values for gt in self.go_traces]
            self.mean_ng_traces = [ng.mean(axis=1).dropna().values for ng in self.ng_traces]


    def make_bold_dfs(self, shape='long', savestr='sim', save=False):
        if not hasattr(self, 'go_traces'):
            self.simulate_bold()
        go_csum = [gt.cumsum(axis=0).max(axis=0).values for gt in self.go_traces]
        ng_csum = [ng.cumsum(axis=0).max(axis=0).values for ng in self.ng_traces]

        if shape == 'fat':
            dfgo_csum = pd.DataFrame.from_dict(OrderedDict(zip(self.labels, go_csum)), orient='index').T
            dfng_csum = pd.DataFrame.from_dict(OrderedDict(zip(self.labels, ng_csum)), orient='index').T
            dfgo_csum.insert(0, 'choice', 'go')
            dfng_csum.insert(0, 'choice', 'nogo')

            self.bold_mag = pd.concat([dfgo_csum, dfng_csum])
            self.bold_mag.index.name = 'cond'
        elif shape == 'long':
            go_csum.extend(ng_csum)
            boldf_list = []
            choices = np.sort(['go', 'nogo'] * self.ncond)
            lbls = [int(lbl) for lbl in self.labels] * 2
            for i, gb in enumerate(go_csum):
                boldf_list.append(pd.DataFrame.from_dict({'cond': lbls[i], 'csum': gb, 'choice': choices[i]}))
            self.bold_mag = pd.concat(boldf_list)

        if save:
            self.bold_mag.to_csv('_'.join([savestr, 'bold_mag.csv']), index=False)


    def cap_n_bound(self, traces, bound, Tg, decay=False):
        """ take 2d array and set values >= upper boundary as np.nan
        return dataframe of capped ndarray, assumes all traces meet
        the criteria trials[any(val >= a)]

        ::Arguments::
              traces (ndarray):
                    go decision traces for single cond
              bound (float):
                    boundary for given condition, all
                    values>=bound --> np.nan
              decay (bool <False>):
                    concatenate mirror image of input array
                    along axis 1 (i.e., bold up, decay down)
        ::Returns::
              traces_df (DataFrame):
                    all go traces
        """

        traces[traces >= bound] = np.nan
        if decay:
            traces_capped = concat((traces, traces[:, ::-1]), axis=1)
        else:
            traces_capped = traces
        traces_df = pd.DataFrame(traces_capped.T)
        traces_df.iloc[Tg:, :] = np.nan
        return traces_df


    def __ix_lastnum__(self, s):
        """ search for index of last numeric (non-NAN)
        value in series (Modified from goo.gl/tId5Rw)

        intended acces through lambda func index_numeric
              df.apply(get_last_numeric, axis=0)
        ::Arguments::
              s (Series):
                    row or column slice of DataFrame
        ::Returns::
              index of last numeric value of series
        """

        if s.last_valid_index() is None:
            return None
        else:
            return s[s.last_valid_index()]


    def __ix_firstnum__(self, s):
        """ see __ix_lastnum__
        ::Arguments::
              s (Series):
                    row or column slice of DataFrame
        ::Returns::
              index of first numeric value of series
        """

        if s.first_valid_index() is None:
            return None
        else:
            return s[s.first_valid_index()]


    def drop_n_roll(self, df, na_position='last'):
        """ rolls all NAN to end of column series for all cols in df
        (Modified from http://goo.gl/3SHD3q)

        ::Arguments:;
              df (DataFrame):
                    accumulators for all ntime (ROWS) x ntrials (COLS)
                    with possible NANs separating inital and
                    re-crossing of threshold
              na_position (str):
                    'first': roll all NANs to top rows
                    'last': roll all NANs to bottom rows
        ::Returns::
              df (DataFrame):
                    accumulators for all ntime (ROWS) x ntrials (COLS)
                    with all NANs rolled to end of row axis
        """

        for c, cseries in df.iteritems():
            result = np.full(len(cseries), np.nan, dtype=cseries.dtype)
            mask = cseries.notnull()
            N = mask.sum()
            if na_position == 'last':
                result[:N] = cseries.loc[mask]
            elif na_position == 'first':
                result[-N:] = cseries.loc[mask]
            else:
                raise ValueError(
                    'na_position {!r} unrecognized'.format(na_position))
            df[c] = result

        return df


    def simulate(self, theta=None, analyze=True):
        """ simulate yhat vector using popt or inits
        if model is not optimized

        :: Arguments ::
              analyze (bool):
                    if True (default) returns yhat vector
                    if False, returns decision traces
        :: Returns ::
              out (array):
                    1d array if analyze is True
                    ndarray of decision traces if False
        """

        theta = self.p
        theta = self.simulator.vectorize_params(theta)
        out = self.simulator.sim_fx(theta, analyze=analyze)
        return out


    def plot_means(self, save=False, ax=None, label_x=True):

        redgreen = lambda nc: sns.blend_palette(["#c0392b", "#27ae60"], n_colors=nc)
        sns.set(style='white', font_scale=1.5)
        if not hasattr(self, 'bold_mag'):
            self.make_bold_dfs()
        if ax is None:
            f, ax = plt.subplots(1, figsize=(5, 5))

        titl = describe_model(self.depends_on)
        df = self.bold_mag.copy()
        df.ix[(df.choice == 'go') & (df.cond <= 50), 'cond'] = 40
        df.ix[(df.choice == 'nogo') & (df.cond >= 50), 'cond'] = 60
        sns.barplot('cond', 'csum', data=df, order=np.sort(df.cond.unique()), palette=redgreen(6), ax=ax)

        mu = df.groupby(['choice', 'cond']).mean()['csum']
        ax.set_ylim(mu.min() * .55, mu.max() * 1.15)
        ax.set_xticklabels([])
        ax.set_xlabel('')
        if label_x:
            ax.set_xlabel('pGo', fontsize=24)
            ax.set_xticklabels(np.sort(df.cond.unique()), fontsize=22)
        ax.set_ylabel('$\Sigma \\theta_{G}$', fontsize=30)
        ax.set_yticklabels([])
        sns.despine()
        plt.tight_layout()
        if save:
            plt.savefig('_'.join([titl, self.decay, 'means.png']), dpi=300)
            plt.savefig(
                '_'.join([titl, self.decay, 'means.svg']), format='svg', rasterized=True)
        return ax


    def plot_traces(self, style='HL', ax=None, label_x=True, save=False):

        cpals = get_cpals()
        sns.set(style='white', font_scale=1.5)
        if ax is None:
            f, ax = plt.subplots(1, figsize=(5, 5))

        tr = self.onset
        titl = describe_model(self.depends_on)
        if style not in ['HL', 'HML']:
            gmu = [ggt.mean(axis=1) for ggt in self.go_traces]
            nmu = [ngt.mean(axis=1) for ngt in self.ng_traces]
        else:
            go_counts = [self.go_traces[i].shape[1]
                         for i in range(len(self.go_traces))]
            ng_counts = [self.ng_traces[i].shape[1]
                         for i in range(len(self.ng_traces))]
            go_lo_ri = randint(0, high=go_counts[0])
            go_hi_ri = randint(0, high=go_counts[-1])
            ng_lo_ri = randint(0, high=ng_counts[0])
            ng_hi_ri = randint(0, high=ng_counts[-1])
            ng_hi_ri = np.argmin(self.ng_traces[-1].iloc[-1, :])

            glow = self.go_traces[0].iloc[:, go_lo_ri].dropna().values
            #pd.concat(self.go_traces[1:3], axis=1).iloc[:,0]
            ghi = self.go_traces[-1].iloc[:, go_hi_ri].dropna().values

            nglow = self.ng_traces[0].iloc[:, ng_lo_ri].dropna().values
            nghi = self.ng_traces[-1].iloc[:, ng_hi_ri].dropna().values

            # return nglow, nghi
            gmu = [glow, ghi]
            nmu = [nglow, nghi]
            tr = [tr[0], tr[-1]]
            gc = ["#40ac5b", '#10ac1d']
            nc = ['#dc3c3c', '#d61b1b']
        #gc = cpals['gpal'](len(gmu))
        #nc = cpals['rpal'](len(nmu))

        gx = [tr[i] + np.arange(len(gmu[i])) * self.dt for i in range(len(gmu))]
        nx = [tr[i] + np.arange(len(nmu[i])) * self.dt for i in range(len(nmu))]
        ls = ['-', '-']
        for i in range(len(gmu)):
            ax.plot(gx[i], gmu[i], linestyle=ls[i], lw=1, color=gc[i])
            ax.fill_between(gx[i], gmu[i], y2=0, lw=1, color=gc[i], alpha=.25)
        for i in range(len(nmu)):
            ax.plot(nx[i], nmu[i], linestyle=ls[i], lw=1, color=nc[i])
            ax.fill_between(nx[i], nmu[i], y2=0, lw=1, color=nc[i], alpha=.25)

        ax.set_ylim(0, self.p['a'].max() * 1.01)
        ax.set_xlim(gx[-1].min() * .98, nx[0].max() * 1.05)
        if label_x:
            ax.set_xlabel('Time', fontsize=26)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_ylabel('$\\theta_{G}$', fontsize=30)

        sns.despine()
        plt.tight_layout()
        if save:
            plt.savefig('_'.join([titl, self.decay, 'traces.png']), dpi=300)
            plt.savefig('_'.join([titl, self.decay, 'traces.svg']), format='svg', rasterized=True)
        return ax



#!usr/bin/env python
from __future__ import division
import pandas as pd
import numpy as np
from scipy.io import loadmat
import os
from copy import deepcopy
from numpy import array
from scipy.stats.distributions import norm, gamma, uniform
from lmfit import Parameters


def random_inits(pkeys, ninits=1, kind='dpm'):
    """ random parameter values for initiating model across range of
    parameter values.
    ::Arguments::
          pkeys (list/dict):
                list of parameter names to sample, makes list of dict keys
                if pkeys is dictionary type
    ::Returns::
          params (dict):
                dictionary with pkeys as keys with values as 1d arrays ninits long
                of randomly sampled parameter values
    """

    if isinstance(pkeys, dict):
        pkeys = pkeys.keys()

    bnd = get_bounds(kind=kind)
    params = {pk: init_distributions(pk, bnd[pk], nrvs=ninits, kind=kind) for pk in pkeys}

    if 'vd' in params.keys():
        params['vi'] = params['vi']*params['vd']

    return params


def loadParameters(inits=None, is_flat=False, kind=None, pc_map={}):
    """ Generates and returns an lmfit Parameters object with
    bounded parameters initialized for flat or non flat model fit
    """

    ParamsObj = Parameters()
    pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso']
    pfit = list(set(inits.keys()).intersection(pnames))
    bounds = get_bounds(kind=kind)
    for pkey, pclist in pc_map.items():
        if is_flat:
            break  # exit
        pfit.remove(pkey)
        if hasattr(inits[pkey], '__iter__'):
            vals = inits[pkey]
        else:
            vals = inits[pkey] * np.ones(len(pclist))
        mn = bounds[pkey][0]
        mx = bounds[pkey][1]
        for k, v in zip(pclist, vals):
            if isinstance(v, np.ndarray):
                v = np.asscalar(v)
            ParamsObj.add(k, value=v, vary=True, min=mn, max=mx)

    for pk in pfit:
        inits = all_params_to_scalar(inits, pfit)
        if is_flat:
            mn = bounds[pk][0]
            mx = bounds[pk][1]
            ParamsObj.add(pk, value=inits[pk], vary=True, min=mn, max=mx)
        else:
            ParamsObj.add(pk, value=inits[pk], vary=False)
    return ParamsObj


def all_params_to_scalar(params, params_list=None, exclude=[]):

    if params_list is None:
        params_list = params.keys()
    for pk in params_list:
        if pk in exclude:
            continue
        if hasattr(params[pk], '__iter__'):
            try:
                params[pk] = np.asscalar(params[pk])
            except ValueError:
                params[pk] = np.mean(params[pk])
    for exc in exclude:
        params[exc] = np.asarray(params[exc], dtype=float)
    return params


def init_distributions(pkey, bounds, tb=.65, kind='dpm', nrvs=25, loc=None, scale=None):
    """ sample random parameter sets to explore global minima (called by
    Optimizer method __hop_around__())
    """
    mu_defaults = {'a': .3, 'tr': .29, 'v': 1., 'vi': .15, 'vd': 1.,
                   'ssv': -1, 'z': .1, 'xb': 1.5, 'sso': .15}
    sigma_defaults = {'a': .15, 'tr': .1, 'v': .15, 'vi': .70, 'vd':.15,
                      'ssv': .15, 'z': .05, 'xb': .5, 'sso': .01}


    if pkey == 'si':
        return .01

    if 'race' in kind or 'iact' in kind:
        mu_defaults['ssv'] = abs(mu_defaults['ssv'])
    if loc is None:
        loc = mu_defaults[pkey]
    if scale is None:
        scale = sigma_defaults[pkey]


    # init and freeze dist shape
    if pkey in ['tr', 'v', 'vd', 'ssv', 'z', 'xb', 'sso']:
        dist = norm(loc, scale)
    elif pkey in ['vi']:
        # vi_perc ~ U(.05, .95) --> vi = vi_perc*vd
        dist = uniform(loc, scale)
    elif pkey in ['a', 'tr']:
        dist = gamma(1, loc, scale)

    # generate random variates
    rvinits = dist.rvs(nrvs)
    while rvinits.min() <= bounds[0]:
        # apply lower limit
        ix = rvinits.argmin()
        rvinits[ix] = dist.rvs()
    while rvinits.max() >= bounds[1]:
        # apply upper limit
        ix = rvinits.argmax()
        rvinits[ix] = dist.rvs()
    return rvinits


def get_bounds(kind='dpm', a=(.1, 1.0), tr=(.1, .54), v=(.1, 5.0), z=(.01, .79), ssv=(-5.0, -.1), xb=(.1, 5), si=(.001, .2), sso=(.01, .3), vd=(.1, 5.0), vi=(.01, .95)):
    """ set and return boundaries to limit search space
    of parameter optimization in <optimize_theta>
    """

    if 'irace' in kind or 'iact' in kind:
        ssv = (abs(ssv[1]), abs(ssv[0]))
    bounds = {'a': a, 'tr': tr, 'v': v, 'ssv': ssv, 'vd':vd, 'vi':vi,
              'z': z, 'xb': xb, 'si': si, 'sso': sso}
    return bounds


def format_local_bounds(xmin, xmax):
    """ groups (xmin, xmax) for each parameter """
    tupler = lambda xlim: tuple([xlim[0], xlim[1]])
    return map((tupler), zip(xmin, xmax))

def format_basinhopping_bounds(basin_keys, nlevels=1):

    allbounds = get_bounds()
    xmin, xmax = [], []
    for pk in basin_keys:
        xmin.append([allbounds[pk][0]] * nlevels)
        xmax.append([allbounds[pk][1]] * nlevels)
    xmin = np.hstack(xmin).tolist()
    xmax = np.hstack(xmax).tolist()
    return xmin, xmax

def get_stepsize_scalars(keys, nlevels=1):

    scalar_dict = {'a': .5, 'tr': .1, 'v': 2., 'vi': 2., 'vd': 2.,
                   'ssv': 2., 'z': .1, 'xb': 1.5, 'sso': .1}

    stepsize_scalars = np.array([scalar_dict[k] for k in keys]*nlevels)
    stepsize_scalars = stepsize_scalars.squeeze()
    return stepsize_scalars


def get_default_inits(kind='dpm', dynamic='hyp', depends_on={}):
    """ if user does not provide inits dict when initializing Model instance,
    grab default dictionary of init params reasonably suited for Model kind
    """

    if 'dpm' in kind:
        popt = {'a': 0.53625, 'v': 1.26, 'xb': 0.87761,
                'ssv': -0.98396, 'tr': 0.17801}
    elif 'pro' in kind:
        popt = {'a': 0.48722, 'xb': 1.51129, 'tr': 0.292126, 'v': 1.718}
    elif 'race' in kind:
        popt = {'a': 0.24266,  'v': 1.05866,  'xb': 1.5,
                'ssv': 1.12441, 'tr': 0.335, 'xb': 1.46335}
    elif 'iact' in kind:
        popt = {'a': 0.44422598, 'sso': 0.2040, 'ssv': 3.02348,
                'tr': 0.21841, 'v': 1.31063, 'xb': 1.46335}
    if 'x' in kind and 'xb' not in popt.keys():
        popt['xb'] = 2.5
    return popt


def check_inits(inits={}, kind='dpm', pdep=[], pro_ss=False, fit_noise=False):
    """ ensure inits dict is appropriate for Model kind
    """
    if 'race' in kind or 'iact' in kind:
        inits['ssv'] = abs(inits['ssv'])
    elif 'dpm' in kind:
        inits['ssv'] = -abs(inits['ssv'])
    if 'pro' in kind:
        if pro_ss and 'ssv' not in inits.keys():
            inits['ssv'] = -0.9976
        elif not pro_ss and 'ssv' in inits.keys():
            ssv = inits.pop('ssv')
    if 'x' in kind and 'xb' not in inits.keys():
        inits['xb'] = 1.5
    if fit_noise or 'si' in pdep and 'si' not in inits.keys():
        inits['si'] = .01
    if 'dpm' not in kind and 'z' in inits.keys():
        discard = inits.pop('z')
    if 'x' not in kind and 'xb' in inits:
        inits.pop('xb')
    return inits


def get_proactive_params(theta, dep='v', pgo=np.arange(0, 120, 20)):
    """ takes pd.Series or dict of params
    and formats for entry to sim
    """
    if not type(theta) == dict:
        theta = theta.to_dict()['mean']
    keep = ['a', 'z', 'v', 'tr', 'ssv', 'ssd']
    keep.pop(keep.index(dep))
    pdict = {pg: theta[dep + str(pg)] for pg in pgo}
    for k in theta.keys():
        if k not in keep:
            theta.pop(k)
    return theta, pdict


def update_params(theta):

    if 't_hi' in theta.keys():
        theta['tr'] = theta['t_lo'] + np.random.uniform() * (theta['t_hi'] - theta['t_lo'])
    if 'z_hi' in theta.keys():
        theta['z'] = theta['z_lo'] + np.random.uniform() * (theta['z_hi'] - theta['z_lo'])
    if 'sv' in theta.keys():
        theta['v'] = theta['sv'] * np.random.randn() + theta['v']

    return theta


def get_intervar_ranges(theta):
    """ theta (dict): parameters dictionary
    """
    if 'st' in theta.keys():
        theta['t_lo'] = theta['tr'] - theta['st'] / 2
        theta['t_hi'] = theta['tr'] + theta['st'] / 2
    if 'sz' in theta.keys():
        theta['z_lo'] = theta['z'] - theta['sz'] / 2
        theta['z_hi'] = theta['z'] + theta['sz'] / 2
    return theta



#!/usr/local/bin/env python
from __future__ import division
import os
import pandas as pd
import numpy as np
from numpy import array
from scipy import optimize
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, messages, analyze
from scipy.stats.mstats import mquantiles as mq
from numpy import cumsum as cs
from numpy import append as app

sns.set(style='white', rc={'text.color': 'black', 'axes.labelcolor': 'black', 'figure.facecolor': 'white'})

cdict = colors.get_cpals('all')
rpal = cdict['rpal']
bpal = cdict['bpal']
gpal = cdict['gpal']
ppal = cdict['ppal']
heat = cdict['heat']
cool = cdict['cool']
slate = cdict['slate']


def scurves(lines=[],  yerr=[], pstop=.5, ax=None, linestyles=None, colors=None, markers=False, labels=None, mc=None, x=None):
    dont_label = False
    sns.set_context('notebook', font_scale=1.7)
    if ax is None:
        f, ax = plt.subplots(1, figsize=(5.5, 6))
    if colors is None:
        colors = slate(len(lines))
    if labels is None:
        labels = [''] * len(lines)
        dont_label = True
    if linestyles is None:
        linestyles = ['-'] * len(lines)

    lines = [(line) if type(line) == list else line for line in lines]
    pse = []
    scl = 100

    if x is None:
        #x = np.arange(len(lines[0]), dtype='float')[::-1]*50
        x = array([400, 350, 300, 250, 200], dtype='float')

    xtls = x.copy()
    xsim = np.linspace(x.min()-20, x.max()+20, 10000)

    yylabel = 'P(Stop)'
    xxlabel = 'SSD (ms)'

    markers = True
    mclinealpha = [.7, 1] * len(lines)

    x = analyze.res(-x, lower=x[0], upper=x[-1])
    xxlim = (x[0]-20, x[-1]+20)
    print(x)
    for i, yi in enumerate(lines):
        color = colors[i]
        y = analyze.res(yi, lower=yi[-1], upper=yi[0])
        p_guess = (np.mean(x), np.mean(y), .5, .2)
        p, cov, infodict, mesg, ier = optimize.leastsq(analyze.residuals, p_guess, args=(x, y), full_output=1, maxfev=10000, ftol=1.e-20)
        x0, y0, c, k = p
        pxp = analyze.sigmoid(p, xsim)
        idx = (np.abs(pxp - pstop)).argmin()
        pse.append(xsim[idx])
        # Plot the results
        if yerr != []:
            ax.errorbar(x, yi, yerr=yerr[i], color=colors[i], lw=0, marker='o', elinewidth=2, ecolor=colors[i])
            ax.plot(xsim, pxp, linestyle=linestyles[i], lw=1.5, color=colors[i], label=labels[i])
            #ax.errorbar(xp, pxp, yerr=yerr, color=color, ecolor=color, capsize=0, lw=0, elinewidth=3)
        if markers:
            a = mclinealpha[i]
            ax.plot(xsim, pxp, linestyle=linestyles[i], lw=2.5, color=color, label=labels[i], alpha=a)
            for ii in range(len(y)):
                ax.plot(x[ii], y[ii], lw=0, marker='o', ms=9, color=color, markerfacecolor='none', mec=color, mew=1.5, alpha=.8)
        else:
            ax.plot(xsim, pxp, linestyle=linestyles[i], lw=3.5, color=colors[i], label=labels[i])
        pse.append(xsim[idx])

    plt.setp(ax, xticks=x, ylim=(-.01, 1.05), yticks=np.arange(0,1.2,.2))#[0,  1])
    #ax.set_xticklabels([int(xt) for xt in xtls])
    #ax.set_yticklabels(np.arange(0,1.2,.2))#[0.0, 1.0])
    #ax.set_xlabel(xxlabel)
    #ax.set_ylabel(yylabel)

    #ax.legend(loc=0)
    plt.tight_layout()
    sns.despine()
    return (pse)


def plot_fits(y, yhat, cdf=True, plot_params={}, save=False, axes=None, kind='', savestr='fit_plot', split='HL', xlim=(.45, .65), label=None, colors=None, data=None, mc=None):

    sns.set_context('notebook', font_scale=1.6)
    pp = plot_params
    if axes is None:
        f, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(14, 5.5), sharey=False)
    else:
        ax1, ax2, ax3 = axes
    if colors is None:
        colors = ["#4168B7"] * 2
    if kind == 'pro':
        xlim = (.48, .58)
    # pull out data vectors
    sc, gq, eq = unpack_yvector(y, kind=kind)
    fitsc, fitgq, fiteq = unpack_yvector(yhat, kind=kind)

    if data is not None:
        axes, pp = plot_data_dists(data, kind=kind, cdf=cdf, axes=[ax1, ax2, ax3], data_type='real')
        fit_cq, fit_eq = [analyze.kde_fit_quantiles(q, bw=.01) for q in [fitgq, fiteq]]
    else:
        kdefits = [analyze.kde_fit_quantiles(q, bw=.01) for q in [gq, fitgq, eq, fiteq]]
        dat_cq, fit_cq, dat_eq, fit_eq = kdefits
        axes, pp = plot_data_dists(data=[dat_cq, dat_eq], kind=kind, cdf=cdf, axes=[ax1, ax2, ax3], data_type='interpolated')

    shade = pp['shade']
    lw = pp['lw']
    ls = pp['ls']
    alpha = pp['alpha']
    bw = pp['bw']
    #sns.distplot(fit_cq, kde=False, color=colors[0], norm_hist=True, ax=ax1, bins=45)
    #sns.distplot(fit_eq, kde=False, color=colors[1], norm_hist=True, ax=ax2, bins=45)
    sns.kdeplot(fit_cq, color='Blue', cumulative=cdf, linestyle='--',
                bw=.01, ax=ax1, linewidth=3, alpha=.70, shade=shade, label=label)
    sns.kdeplot(fit_eq, color='Red', cumulative=cdf, linestyle='--',
                bw=.01, ax=ax2, linewidth=3, alpha=.60, shade=shade)

    for ax in axes:
        if ax.is_last_col():
            continue
        ax.set_xlim(xlim[0], xlim[1])
        if ax.is_first_col():
            ax.set_ylabel('P(RT)')
        if ax.is_last_row():
            ax.set_xlabel('RT (ms)')
        ax.set_xticklabels([int(xx) for xx in ax.get_xticks() * 1000])

    # Plot observed and predicted stop curves
    scurves([sc, fitsc], kind=kind, linestyles=['-', '--'], ax=ax3, colors=colors, markers=True, mc=mc)
    plt.tight_layout()
    sns.despine()
    if save:
        plt.savefig(savestr + '.png', format='png', dpi=300)


def react_fit_plots(m, color="#4168B7", is_flat=False, save=False):

    #redata = m.data
    sns.set_context('notebook', font_scale=1.6)
    if is_flat:
        f, axes = plt.subplots(1, 3, figsize=(12, 4))
        axes = array([axes])
        fits = [m.fits[0]]
        y = [m.flat_y]
        ncond = 1
    else:
        f, axes = plt.subplots(2, 3, figsize=(12, 7))
        fits = m.fits.reshape(2, 16)
        y = m.avg_y
        ncond = m.ncond

    labels = ['Baseline', 'Caution']

    for i in range(ncond):
        plot_fits(y[i], fits[i], kind='', colors=[color] * 2, axes=axes[i])

    for ax in axes.flatten():
        if ax.is_last_col():
            continue
        ax.set_xlim(.4, .65)
        if ax.is_first_col():
            ax.set_ylabel('P(RT)')
        if ax.is_last_row():
            ax.set_xlabel('RT (ms)')
        ax.set_xticklabels([int(xx) for xx in ax.get_xticks() * 1000])

    if save:
        kind = m.kind
        mdescr = messages.describe_model(m.depends_on)
        if 'and' in mdescr:
            mdeps = mdescr.split(' and ')
            mdescr = '_'.join([dep for dep in mdeps])

        savestr = '_'.join([kind, mdescr, 'fits'])
        plt.savefig(savestr + '.png', dpi=500)
        plt.savefig(savestr + '.svg', rasterized=True)


def profits(y, yhat, cdf=False, plot_params={}, save=False, axes=None, kind='', savestr='fit_plot', split='HL', xlim=(.4, .65), label=None, colors=None, data=None, mc=None):
    sns.set_context('notebook', font_scale=1.6)

    pp = plot_params
    if axes is None:
        f, (ax1, ax2, ax3) = plt.subplots(
            1, 3, figsize=(14, 5.5), sharey=False)
    else:
        ax1, ax2, ax3 = axes
    if colors is None:
        colors = ["#4168B7"] * 2
    if kind == 'pro':
        xlim = (.48, .58)
    # pull out data vectors
    sc, gq, eq = unpack_yvector(y, kind=kind)
    fitsc, fitgq, fiteq = unpack_yvector(yhat, kind=kind)
    kdefits = [analyze.kde_fit_quantiles(q, bw=.01) for q in [gq, fitgq, eq, fiteq]]
    dat_cq, fit_cq, dat_eq, fit_eq = kdefits
    sns.distplot(fit_cq, kde=False, color=colors[0], norm_hist=True, ax=ax1, bins=25)
    sns.distplot(fit_eq, kde=False, color=colors[1], norm_hist=True, ax=ax2, bins=25)

    for ax in axes:
        if ax.is_last_col():
            continue
        ax.set_xlim(xlim[0], xlim[1])
        if ax.is_first_col():
            ax.set_ylabel('P(RT)')
        if ax.is_last_row():
            ax.set_xlabel('RT (ms)')
        ax.set_xticklabels([int(xx) for xx in ax.get_xticks() * 1000])

    # Plot observed and predicted stop curves
    scurves([sc, fitsc], kind=kind, linestyles=['-', '--'], ax=ax3, colors=colors, markers=True, mc=mc)
    plt.tight_layout()
    sns.despine()
    if save:
        plt.savefig(savestr + '.png', format='png', dpi=300)


def plot_data_dists(data, kind='', data_type='real', cdf=False, axes=[], get_rts=False):

    emp_kq = lambda rts: analyze.kde_fit_quantiles(mq(rts, prob=np.arange(0, 1, .02)), bw=.008)

    ax1, ax2, ax3 = axes
    if data_type == 'real':
        if kind == 'pro':
            hi_rts = data.query('response==1 & pGo>.5').rt.values
            lo_rts = data.query('response==1 & pGo<.5').rt.values
        elif kind == '':
            hi_rts = data.query('response==1 & acc==1').rt.values
            lo_rts = data.query('response==1 & acc==0').rt.values
        dat_cq = emp_kq(hi_rts)
        dat_eq = emp_kq(lo_rts)
        if get_rts:
            return axes, plot_params, rts
    elif data_type == 'interpolated':
        dat_cq, dat_eq = data

    if cdf:
        shade = False
        alpha = 1
        bw = .01
        lw = 3.5
        ls = '--'
        sns.kdeplot(dat_cq, color='k', cumulative=cdf, ax=ax1, linewidth=lw, linestyle='-')
        sns.kdeplot(dat_eq, color='k', cumulative=cdf, ax=ax2, linewidth=lw, linestyle='-')
    else:
        # set parameters for simulated plots
        shade = True
        alpha = .5
        bw = .01
        lw = 2.5
        ls = '-'
        sns.distplot(dat_cq, kde=False, color='k', norm_hist=True, ax=ax1, bins=50)
        sns.distplot(dat_eq, kde=False, color='k', norm_hist=True, ax=ax2, bins=50)

    plot_params = {'shade': shade, 'alpha': alpha, 'bw': bw, 'lw': lw, 'ls': ls}
    if get_rts:
        return axes, plot_params, rts
    return axes, plot_params


def plot_reactive_fits(model, cumulative=True, save=False, col=None):

    sns.set_context('notebook', font_scale=1.6)
    f, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(14, 5))
    y = model.avg_y
    r, c = y.shape
    if col is None:
        col = [bpal(2), rpal(2)]
    yhat = model.fits.reshape(r, c)
    yh_id = 'fits.png'

    if save:
        savestr = '_'.join([get_model_name(model), yh_id])

    linestyles = ['-', '--']
    for i in range(model.ncond):
        sc, cq, eq = unpack_yvector(y[i])
        sc_hat, cqhat, eqhat = unpack_yvector(yhat[i])

        for ii, qc in enumerate([cq, cqhat]):
            kdeqc = analyze.kde_fit_quantiles(qc, bw=.01)
            sns.kdeplot(kdeqc, linestyle=linestyles[ii], cumulative=cumulative, ax=ax1, linewidth=3.5, color=col[i][ii])
        for ii, qe in enumerate([eq, eqhat]):
            kdeqe = analyze.kde_fit_quantiles(qe, bw=.01)
            sns.kdeplot(kdeqe, cumulative=cumulative, linestyle=linestyles[ii], ax=ax2, linewidth=3.5, color=col[i][ii])

        labels = [' '.join([model.labels[i], x]) for x in ['data', 'model']]
        # Plot observed and predicted stop curves
        scurves([sc, sc_hat], labels=labels, kind='', colors=col[i], linestyles=['-', '--'], ax=ax3, markers=True)
        plt.tight_layout()
        sns.despine()

    ax1.set_title('Correct RTs', fontsize=17)
    ax2.set_title('SS Trial RTs (Errors)', fontsize=17)
    ax3.set_title('P(Stop) Across SSD', fontsize=17)
    #[axx.set_xlim(.470,.650) for]

    for axx in [ax1, ax2]:
        axx.set_xlim(.470, .650)
        xxticks = axx.get_xticks()
        axx.set_ylabel('P(RT<t)')
        axx.set_xlabel('RT (s)')
        #axx.set_ylim(-.05, 1.05)
        axx.set_xticks(xxticks)
        axx.set_xticklabels([int(xx) for xx in xxticks * 1000])
        axx.set_xlim(.470, .650)
    if save:
        plt.savefig(savestr, dpi=300)


def unpack_yvector(y, kind=''):

    if 'pro' in kind:
        sc, gq, eq = y[:6], y[6:11], y[11:]
    else:
        sc, gq, eq = y[1:6], y[6:11], y[11:]

    return sc, gq, eq


def get_model_name(model):
    mname = model.kind
    mdep = messages.describe_model(model.depends_on)
    if 'x' in model.kind:
        mname = '_'.join([mname, model.dynamic])
    mname = '_'.join([mname, mdep])
    return mname


def plot_idx_fits(obs, sim, kind='', save=False):

    if kind == '':
        df = df.where(df > 0).dropna()
        for idx, idx_c in obs.iterrows():
            try:
                save_str = '_'.join([str(idx), idx_c['Cond'], 'pred'])
                y = idx_c.loc['Go':'e90'].values.astype(np.float)
                yhat = df.iloc[idx, :].values.astype(np.float)
                plot_fits(y, yhat, kind='', save=save, savestr=save_str)
            except Exception:
                continue
    elif kind == 'pro':
        """
        TODO
        """

        df = None


def plot_kde_cdf(quant, bw=.1, ax=None, color=None):

    if ax is None:
        f, ax = plt.subplots(1, figsize=(5, 5))
    if color is None:
        color = 'k'
    kdefits = analyze.kde_fit_quantiles(quant, bw=bw)
    sns.kdeplot(kdefits, cumulative=True,  color=color, ax=ax, linewidth=2.5)

    ax.set_xlim(kdefits.min() * .94, kdefits.max() * 1.05)
    ax.set_ylabel('P(RT<t)')
    ax.set_xlabel('RT (s)')
    ax.set_ylim(-.05, 1.05)
    ax.set_xticklabels(ax.get_xticks() * .1)

    plt.tight_layout()
    sns.despine()


def gen_pro_traces(ptheta, bias_vals=[], bias='v', integrate_exec_ss=False, return_exec_ss=False, pgo=np.arange(0, 1.2, .2)):

    dvglist = []
    dvslist = []

    if bias_vals == []:
        deplist = np.ones_like(pgo)

    for val, pg in zip(bias_vals, pgo):
        ptheta[bias] = val
        ptheta['pGo'] = pg
        dvg, dvs = RADD.run(ptheta, ntrials=10, tb=.565)
        dvglist.append(dvg[0])

    if pg < .9:
        dvslist.append(dvs[0])
    else:
        dvslist.append([0])

    if integrate_exec_ss:
        ssn = len(dvslist[0])
        traces = [np.append(dvglist[i][:-ssn], (dvglist[i][-ssn:] + ss) - dvglist[i][-ssn:]) for i, ss in enumerate(dvslist)]
        traces.append(dvglist[-1])
        return traces

    elif return_exec_ss:
        return [dvglist, dvslist]
    else:
        return dvglist


def gen_re_traces(model, params, integrate_exec_ss=False, ssdlist=np.arange(.2, .45, .05), integrate=False):

    nssd = model.simulator.nssd
    nss = model.simulator.nss
    nsstot = nss * nssd
    nc = model.simulator.ncond

    dvg, dvs = model.simulate(params, analyze=False, return_traces=True)
    gtraces = dvg[0, :nssd]
    straces = [dvs[0, i, i] for i in range(nssd)]

    Ps, Ts = model.simulator.__update_stop_process__(params)
    Pg, Tg = model.simulator.__update_go_process__(params)
    bound = params['a'][0]
    gtraces = [gt[gt < bound] for gt in gtraces]
    ssi, xinit_ss, integrated, dvgs, dvss = [], [], [], [], []
    for i, (g, s) in enumerate(zip(gtraces, straces)):
        xinit_ss.append(Tg[0] - Ts[i])
        ssi.append(g[:xinit_ss[i]])
        ss = s[:Ts[i]]
        s = np.append(g[:xinit_ss[i]], ss[ss >= 0])
        ixmin = np.min([len(g), len(s)])
        dvgs.append(g[:ixmin])
        dvss.append(s[:ixmin])
        if integrate:
            tx = xinit_ss[i]
            integrated.append(app(g[:tx], cs(app(g[tx], (np.diff(g[tx:]) + np.diff(s[tx:]))))))

    nframes = [len(gt) for gt in dvgs]
    x = params['tr'][0] * 1000 + [np.arange(nf) for nf in nframes]
    return [x, dvgs, dvss, xinit_ss, ssi, np.max(nframes)]


def build_decision_axis(onset, bound, ssd=np.arange(200, 450, 50), tb=650):

    # init figure, axes, properties
    f, axes = plt.subplots(len(ssd), 1, figsize=(5, 5), dpi=300)
    #f.subplots_adjust(wspace=.05, top=.1, bottom=.1)
    f.subplots_adjust(hspace=.05, top=.99, bottom=.05)
    w = tb + 40
    h = bound
    start = onset - 80
    # c=["#e74c3c", '#27ae60', '#4168B7', '#8E44AD']
    for i, ax in enumerate(axes):
        plt.setp(ax, xlim=(start - 1, w + 1), ylim=(0 - (.01 * h), h + (.01 * h)))
        ax.vlines(x=ssd[i], ymin=0, ymax=h, color="#e74c3c", lw=1.5, alpha=.5)
        ax.hlines(y=h, xmin=start, xmax=w, color='k')
        ax.hlines(y=0, xmin=start, xmax=w, color='k')
        ax.vlines(x=tb, ymin=0, ymax=h, color='#2043B0', lw=1.5, linestyle='-', alpha=.5)
        ax.vlines(x=start + 2, ymin=0, ymax=h, color='k')
        ax.text(ssd[i] + 10, h * .87, str(ssd[i]) + 'ms', fontsize=15)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
    sns.despine(top=True, right=True, bottom=True, left=True)

    return f, axes

def build_multi_axis(p, nresp=4, tb=1000):
    bound = p['a']
    onset = p['tr']
    if hasattr(bound, '__iter__'):
        bound = bound[0]
        onset = onset[0]
    # init figure, axes, properties
    f, axes = plt.subplots(2, 2, figsize=(10, 5), sharex=True, sharey=True)
    f.subplots_adjust(hspace=.1, top=.99, bottom=.05)
    w = tb + 40
    h = bound
    start = onset - 80
    axes=axes.flatten()
    # c=["#e74c3c", '#27ae60', '#4168B7', '#8E44AD']
    for i, ax in enumerate(axes):
        plt.setp(ax, xlim=(start - 1, w + 1), ylim=(0 - (.01 * h), h + (.01 * h)))
        ax.hlines(y=h, xmin=start, xmax=w, color='k')
        ax.hlines(y=0, xmin=start, xmax=w, color='k')
        ax.vlines(x=tb, ymin=0, ymax=h, color='#2043B0', lw=1.5, linestyle='-', alpha=.5)
        ax.vlines(x=start + 2, ymin=0, ymax=h, color='k')
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
    sns.despine(top=True, right=True, bottom=True, left=True)
    return f, axes


def re_animate_multiax(i, x, gtraces, glines, straces, slines, params, xi, yi):

    gcolor = '#27ae60'
    scolor = '#e74c3c'
    for n in range(len(x)):
        ex, gt, gl, st, sl, ix, iy = [xx[n] for xx in [x, gtraces, glines, straces, slines, xi, yi]]
        try:
            gl.set_data(ex[:i + 1], gt[:i + 1])
            gl.set_color(gcolor)
            sl.set_data(ex[ix:i + 1], st[ix:i + 1])
            sl.set_color(scolor)
        except Exception:
            return sl, gl
        #f.savefig('animation_frames/movie/img' + str(i) +'.png', dpi=300)
    return sl, gl


def pro_animate(i, x, protraces, prolines):

    clist = sns.color_palette('autumn', n_colors=6)[::-1]

    for nline, (pline, ptrace) in enumerate(zip(prolines, protraces)):
        pline.set_data(x[:i + 1], ptrace[:i + 1])
        pline.set_color(clist[nline])

    return prolines,


def anim_to_html(anim):

    from tempfile import NamedTemporaryFile

    VIDEO_TAG = """<video controls>
             <source src="data:video/x-m4v;base64,{0}" type="video/mp4">
             Your browser does not support the video tag.
            </video>"""

    if not hasattr(anim, '_encoded_video'):
        with NamedTemporaryFile(suffix='.mp4') as f:
            anim.save(f.name, dpi=300, extra_args=['-vcodec', 'libx264', '-pix_fmt', 'yuv420p'])
            video = open(f.name, "rb").read()
        anim._encoded_video = video.encode("base64")
    return VIDEO_TAG.format(anim._encoded_video)


def display_animation(anim):
    from IPython.display import HTML
    plt.close(anim._fig)
    return HTML(anim_to_html(anim))


def plot_all_traces(DVg, DVs, theta, ssd=np.arange(.2, .45, .05), kind='dpm'):

    ncond = DVg.shape[0]
    nssd = DVs.shape[1]
    f, axes = plt.subplots(nssd, ncond, figsize=(12, 14))
    for i in range(ncond):
        params = {k: v[i] if hasattr(v, '__iter__') else v for k, v in theta.items()}
        for ii in range(nssd):
            plot_traces(DVg=DVg[i], DVs=DVs[i, ii], ssd=ssd[ii], sim_theta=params, ax=axes[ii, i], kind=kind)
    return f


def plot_traces(DVg=[], DVs=[], sim_theta={}, kind='dpm', ssd=.450, ax=None, tau=.005, tb=.650, cg='#16a085', cr=['#f1c40f', '#e74c3c', '#c0392b']):

    if ax is None:
        f, ax = plt.subplots(1, figsize=(8, 5))
    tr = sim_theta['tr']
    a = sim_theta['a']
    z = 0
    for i, igo in enumerate(DVg):
        ind = np.argmax(igo >= a)
        xx = [np.arange(tr, tr + (len(igo[:ind - 1]) * tau), tau), np.arange(tr, tb, tau)]
        x = xx[0 if len(xx[0]) < len(xx[1]) else 1]
        plt.plot(x, igo[:len(x)], color=cg, alpha=1, linewidth=2)

        ax.vlines(ssd, 0, a, linewidth=1, alpha=.5, color='#6C7A89')
        if kind in ['irace', 'dpm', 'iact'] and i < len(DVs[0]):
            if np.any(DVs[0] <= 0):
                ind = np.argmax(DVs[0] <= 0)
            else:
                ind = np.argmax(DVs[0] >= a)
            xx = [np.arange(ssd, ssd + (len(DVs[0][:ind - 1]) * tau), tau), np.arange(ssd, tb, tau)]
            x = xx[0 if len(xx[0]) < len(xx[1]) else 1]
            crx = sns.blend_palette(cr, n_colors=len(x))

            for ii in range(len(x)):
                if ii == len(x) - 1:
                    break
                plt.plot([x[ii], x[ii + 1]], [np.asscalar(DVs[0][ii]), np.asscalar(DVs[0][ii + 1])], color=crx[ii], alpha=.9, linewidth=2)

    xlow = np.min([tr, ssd])
    xlim = (xlow * .75, 1.05 * tb)

    print xlim
    if kind == 'pro' or np.any(DVs <= 0):
        ylow = 0
        ylim = (-.03, a * 1.03)
    else:
        ylow = z
        ylim = (z - .03, a * 1.03)

    plt.setp(ax, xlim=xlim, ylim=ylim)
    ax.hlines(y=z, xmin=xlow, xmax=tb, linewidth=2, linestyle='--', color="k", alpha=.5)
    ax.hlines(y=a, xmin=xlow, xmax=tb, linewidth=2, linestyle='-', color="k")
    ax.hlines(y=ylow, xmin=xlow, xmax=tb, linewidth=2, linestyle='-', color="k")
    ax.vlines(x=xlow, ymin=ylow * .998, ymax=a * 1.002, linewidth=2, linestyle='-', color="k")
    sns.despine(top=True, bottom=True, right=True, left=True)
    ax.set_xlim(xlim)
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_xticks([])
    ax.set_yticks([])

    return ax



#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import hstack as hs
from numpy import newaxis as na
from numpy import cumsum as csum
from scipy.stats.mstats import mquantiles as mq

class Simulator(object):
    """ Core code for simulating models. All cond, trials, &
    timepoints are simulated simultaneously
    """

    def __init__(self, fitparams=None, inits=None, pc_map=None, kind='xdpm', dt=.001, si=.01, base=0, learn=False):

        self.dt = dt
        self.si = si
        self.dx = np.sqrt(self.si * self.dt)
        self.base = base

        self.learn = learn

        self.inits = inits
        self.kind = kind
        self.pc_map = pc_map

        self.__update__(fitparams=fitparams)
        self.__init_model_functions__()
        self.__init_analyze_functions__()


    def __update__(self, fitparams=None):

        if fitparams:
            fp = dict(deepcopy(fitparams))
            self.fitparams = fitparams
        else:
            fp = dict(deepcopy(self.fitparams))

        # set y, wts to be used cost function
        self.y = self.fitparams['y'].flatten()
        self.wts = self.fitparams['wts'].flatten()

        # misc data & model parameters used during
        # simulations/optimizations
        self.nlevels = self.fitparams['y'].ndim
        self.tb = fp['tb']
        self.ntot = fp['ntrials']
        self.percentiles = fp['percentiles']
        self.dynamic = fp['dynamic']

        # include SSD's if stop-signal task
        if 'ssd' in list(fp):
            self.ssd = fp['ssd']
            self.nssd = fp['nssd']
            self.nss = int((.5 * self.ntot) / self.nssd)

        # non conditional parameters
        self.pvc = ['a', 'tr', 'v', 'xb', 'ssv']
        if self.nlevels>1:
            # remove any parameters free to vary across experimental conditions
            map((lambda pkey: self.pvc.remove(pkey)), list(self.pc_map))


    def __prep_global__(self, basin_params={}, basin_keys=[]):
        # set
        self.basin_keys = basin_keys
        # set all constant parameters in
        # basin_params object (not available
        # to basinhopping minimizer)
        self.basin_params = basin_params
        self.chunk = lambda x, n: [array(x[i:i + n]) for i in range(0, len(x), n)]


    def __init_model_functions__(self):
        """ initiates the simulation function used in
        optimization routine
        """

        if 'dpm' in self.kind:
            self.sim_fx = self.simulate_dpm
            self.analyze_fx = self.analyze_reactive
        elif 'pro' in self.kind:
            self.sim_fx = self.simulate_pro
            self.analyze_fx = self.analyze_proactive
        elif 'irace' in self.kind:
            self.sim_fx = self.simulate_irace
            self.analyze_fx = self.analyze_reactive
        elif 'iact' in self.kind:
            self.sim_fx = self.simulate_interactive
            self.analyze_fx = self.analyze_interactive
        elif 'nalt' in self.kind:
            self.simfx = self.simulate_nalt
            self.analyze_fx = self.analyze_nalt
        if 'rl' in self.kind:
            self.trialwise = True


        # SET STATIC/DYNAMIC BASIS FUNCTIONS
        if 'x' in self.kind and self.dynamic == 'hyp':
            # dynamic bias is hyperbolic
            self.temporal_dynamics = lambda p, t: np.cosh(p['xb'][:, na] * t)
        elif 'x' in self.kind and self.dynamic == 'exp':
            # dynamic bias is exponential
            self.temporal_dynamics = lambda p, t: np.exp(p['xb'][:, na] * t)
        else:
            self.temporal_dynamics = lambda p, t: np.ones((self.nlevels, len(t)))

    def __init_analyze_functions__(self):
        """ initiates the analysis function used in
        optimization routine to produce the yhat vector
        """

        prob = self.percentiles
        dt = self.dt


        if self.learn:
            self.go_resp = lambda trace, a: np.argmax((trace.T >= a).T, axis=1) * dt
            self.ss_resp_up = lambda trace, a: np.argmax((trace.T >= a).T, axis=2) * dt
            self.ss_resp_lo = lambda trace, x: np.argmax((trace.T <= 0).T, axis=2) * dt
            self.RT = lambda ontime, rbool: ontime[:, na] + (rbool * np.where(rbool == 0, np.nan, 0))
            self.RTQ = lambda zpd: map((lambda x: mq(x[0][x[0] < x[1]], prob)), zpd)
        else:
            self.go_resp = lambda trace, a: np.argmax((trace.T >= a).T, axis=2) * dt
            self.ss_resp_up = lambda trace, a: np.argmax((trace.T >= a).T, axis=3) * dt
            self.ss_resp_lo = lambda trace, x: np.argmax((trace.T <= 0).T, axis=3) * dt
            self.RT = lambda ontime, rbool: ontime[:, na] + (rbool * np.where(rbool == 0, np.nan, 1))
            self.RTQ = lambda zpd: map((lambda x: mq(x[0][x[0] < x[1]], prob)), zpd)

        if 'irace' in self.kind:
            self.ss_resp = self.ss_resp_up
        else:
            self.ss_resp = self.ss_resp_lo



    def vectorize_params(self, p):
        """ ensures that all parameters are converted to arrays before simulation. see
        doc strings for prepare_fit() method of Model class (in build.py) for details
        regarding pc_map and logic for fitting models with parameters that depend on
        experimental conditions

        ::Arguments::
              p (dict):
                    keys are parameter names (e.g. ['v', 'a', 'tr' ... ])
                    values are parameter values, can be vectors or floats
        ::Returns::
              p (dict):
                    dictionary with all parameters as vectors
        """

        if 'si' in list(p):
             self.dx = np.sqrt(p['si'] * self.dt)
        if 'xb' not in list(p):
            p['xb'] = 1.0
        for pkey in self.pvc:
            p[pkey] = p[pkey] * np.ones(self.nlevels).astype(np.float32)
        for pkey, pkc in self.pc_map.items():
            if self.nlevels == 1:
                break
            elif pkc[0] not in list(p):
                p[pkey] = p[pkey] * np.ones(len(pkc)).astype(np.float32)
            else:
                p[pkey] = array([p[pc] for pc in pkc]).astype(np.float32)
        return p


    def basinhopping_minimizer(self, x):
        """ used specifically by fit.perform_basinhopping() for Model
        objects with multiopt attr (See __opt_routine__ and perform_basinhopping
        methods of Optimizer object)
        """

        p = dict(deepcopy(self.basin_params))
        # segment 'x' into equal len arrays (one array,
        # nlevels vals long per free parameter) in basin_keys
        px = self.chunk(x, self.nlevels)

        for i, pk in enumerate(self.basin_keys):
            p[pk] = px[i]

        # simulate using filled params dictionary
        yhat = self.sim_fx(p)

        # calculate and return cost error
        return np.sum(self.wts * (yhat - self.y)**2)


    def __cost_fx__(self, theta):
        """ Main cost function used for fitting all models self.sim_fx
        determines which model is simulated (determined when Simulator
        is initiated)
        """
        if type(theta) == dict:
            p = dict(deepcopy(theta))
        else:
            p = theta.valuesdict()

        yhat = self.sim_fx(p, analyze=True)
        return np.sum(self.wts * (self.y - yhat)**2).astype(np.float32)


    def __update_go_process__(self, p):
        """ update Pg (probability of DVg +dx) and Tg (num go process timepoints)
        for go process and get get dynamic bias signal if 'x' model
        """
        Pg = 0.5 * (1 + p['v'] * self.dx / self.si)
        Tg = np.ceil((self.tb - p['tr']) / self.dt).astype(int)
        t = csum([self.dt] * Tg.max())
        self.xtb = self.temporal_dynamics(p, t)
        #diff = [Tg.max()-tg for tg in Tg]
        #self.xtb = np.vstack([np.append(np.ones(diff[i]), xtb[i][:tg]) for i, tg in enumerate(Tg)])
        return Pg, Tg


    def __update_stop_process__(self, p, sso=0):
        """ update Ps (probability of DVs +dx) and
        Ts (SSD onsets) for stop process
        """
        if 'sso' in list(p):
            sso = p['sso']
        Ps = 0.5 * (1 + p['ssv'][0] * self.dx / self.si)
        Ts = np.ceil((self.tb - (self.ssd + sso)) / self.dt).astype(int)
        return Ps, Ts


    def __update_interactive_params__(self, p):
        # add ss interact delay to SSD
        Ps, Ts = self.__update_stop_process__(p)
        Pg = 0.5 * (1 + p['v'] * self.dx / self.si)
        Tg = np.ceil((self.tb - p['tr']) / self.dt).astype(int)
        nt = np.max(np.hstack([Tg, Ts]))
        t = csum([self.dt] * nt)
        self.xtb = self.temporal_dynamics(p, t)
        return Pg, Tg, Ps, Ts, nt


    def simulate_dpm(self, p, analyze=True):
        """ Simulate the dependent process model (DPM)
        ::Arguments::
            p (dict):
                parameter dictionary. values can be single floats
                or vectors where each element is the value of that
                parameter for a given condition
            analyze (bool <True>):
                if True (default) return rt and accuracy information
                (yhat in cost fx). If False, return Go and Stop proc.
        ::Returns::
            yhat of cost vector (ndarray)
            or Go & Stop processes in list (list of ndarrays)
        """

        p = self.vectorize_params(p)
        Pg, Tg = self.__update_go_process__(p)
        Ps, Ts = self.__update_stop_process__(p)

        ntot = self.ntot
        nss = self.nss

        nl = self.nlevels
        dx = self.dx

        DVg = self.xtb[:, na] * csum(np.where((rs((nl, ntot, Tg.max())).T < Pg), dx, -dx).T, axis=2)
        # INITIALIZE DVs FROM DVg(t=SSD)
        init_ss = array([[DVg[i, :nss, ix] for ix in np.where(Ts < Tg[i], Tg[i] - Ts, 0)] for i in range(nl)])
        DVs = init_ss[:, :, :, na] + csum(np.where(rs((nss, Ts.max())) < Ps, dx, -dx), axis=1)
        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        else:
            return [DVg, DVs]


    def simulate_pro(self, p, analyze=True):
        """ Simulate the proactive competition model
        (see simulate_dpm() for I/O details)
        """
        p = self.vectorize_params(p)
        Pg, Tg = self.__update_go_process__(p)
        nl = self.nlevels
        dx = self.dx
        ntot = self.ntot

        DVg = self.xtb[:, na] * csum(np.where((rs((nl, ntot, Tg.max())).T < Pg), dx, -dx).T, axis=2)
        if analyze:
            return self.analyze_fx(DVg, p)
        return DVg


    def simulate_irace(self, p, analyze=True):
        """ simulate the independent race model
        (see simulate_dpm() for I/O details)
        """
        p = self.vectorize_params(p)
        Pg, Tg = self.__update_go_process__(p)
        Ps, Ts = self.__update_stop_process__(p)

        nss = self.nss
        ntot = self.ntot

        nssd = self.nssd
        nl = self.nlevels
        dx = self.dx

        DVg = self.xtb[:, na] * csum(np.where((rs((nl, ntot, Tg.max())).T < Pg), dx, -dx).T, axis=2)
        # INITIALIZE DVs FROM 0
        DVs = csum(np.where(rs((nl, nssd, nss, Ts.max())) < Ps, dx, -dx), axis=3)
        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        return [DVg, DVs]


    def simulate_rldpm(self, p, analyze=True):
        """ Simulate the dependent process model (DPM)
        ::Arguments::
            p (dict):
                parameter dictionary. values can be single floats
                or vectors where each element is the value of that
                parameter for a given condition
            analyze (bool <True>):
                if True (default) return rt and accuracy information
                (yhat in cost fx). If False, return Go and Stop proc.
        ::Returns::
            yhat of cost vector (ndarray)
            or Go & Stop processes in list (list of ndarrays)
        """

        p = self.vectorize_params(p)
        Pg, Tg = self.__update_go_process__(p)
        Ps, Ts = self.__update_stop_process__(p)

        ntot = self.ntot
        nss = self.nss
        nl = self.nlevels
        dx = self.dx

        for trial in range(ntot):
            DVg = self.xtb[:, na] * csum(np.where((rs((nl, Tg.max())).T<Pg), dx, -dx).T, axis=1)
            # INITIALIZE DVs FROM DVg(t=SSD)
            if trial%2:
                init_ss = array([[DVg[i, ix] for ix in np.where(Ts<Tg[i], Tg[i]-Ts, 0)] for i in range(nl)])
                DVs = init_ss[:,:,na] + csum(np.where(rs((nss, Ts.max()))<Ps, dx, -dx), axis=1)

        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        else:
            return [DVg, DVs]


    def analyze_rldpm(self, DVg, DVs=None, p=None):
        """ get rt and accuracy of go and stop process for simulated
        conditions generated from simulate_dpm
        """

        prob = self.percentiles
        ssd = self.ssd
        tb = self.tb
        nl = self.nlevels
        nssd = self.nssd

        gdec = self.go_resp(DVg, p['a'])
        gort = self.RT(p['tr'], gdec)

        sdec = self.ss_resp(DVs, p['a'])
        ssrt = self.RT(ssd, sdec)
        ert = np.tile(gort, nssd).reshape(nl, nssd)

        eq = self.RTQ(zip(ert, ssrt))
        gq = self.RTQ(zip(gort, [tb] * nl))
        gacc = np.nanmean(np.where(gort < tb, 1, 0), axis=1)
        sacc = np.where(ert < ssrt, 0, 1).mean(axis=2)
        return hs([hs([i[ii] for i in [gacc, sacc, gq, eq]]) for ii in range(nl)])


    def analyze_reactive(self, DVg, DVs, p):
        """ get rt and accuracy of go and stop process for simulated
        conditions generated from simulate_dpm
        """
        nss = self.nss
        prob = self.percentiles
        ssd = self.ssd
        tb = self.tb
        nl = self.nlevels
        nssd = self.nssd

        gdec = self.go_resp(DVg, p['a'])
        sdec = self.ss_resp(DVs, p['a'])

        gort = self.RT(p['tr'], gdec)
        ssrt = self.RT(ssd, sdec)
        ert = np.tile(gort[:, :nss], nssd).reshape(nl, nssd, nss)

        eq = self.RTQ(zip(ert, ssrt))
        gq = self.RTQ(zip(gort, [tb] * nl))
        gacc = np.nanmean(np.where(gort < tb, 1, 0), axis=1)
        sacc = np.where(ert < ssrt, 0, 1).mean(axis=2)
        return hs([hs([i[ii] for i in [gacc, sacc, gq, eq]]) for ii in range(nl)])


    def analyze_proactive(self, DVg, p):
        """ get proactive rt and accuracy of go process for simulated
        conditions generated from simulate_pro
        """
        prob = self.percentiles
        ssd = self.ssd
        tb = self.tb
        nl = self.nlevels

        gdec = self.go_resp(DVg, p['a'])
        gort = self.RT(p['tr'], gdec)
        gq = hs(self.RTQ(zip(gort, [tb] * nl)))

        # Get response and stop accuracy information
        gacc = 1 - np.mean(np.where(gort < tb, 1, 0), axis=1)
        return hs([gacc, gq])


    def mean_pgo_rts(self, p, return_vals=True):
        """ Simulate proactive model and calculate mean RTs
        for all conditions rather than collapse across high and low
        """
        import pandas as pd

        DVg = self.simulate_pro(p, analyze=False)
        gdec = self.go_resp(DVg, p['a'])

        rt = self.RT(p['tr'], gdec)
        mu = np.nanmean(rt, axis=1)
        ci = pd.DataFrame(rt.T).sem() * 1.96
        std = pd.DataFrame(rt.T).std()

        self.pgo_rts = {'mu': mu, 'ci': ci, 'std': std}
        if return_vals:
            return self.pgo_rts


    def analyze_pro_data(self, DVg, p):
        prob = self.percentiles
        tb = self.tb
        nl = self.nlevels
        ix = self.rt_cix
        gdec = self.go_resp(DVg, p['a'])
        rt = self.RT(p['tr'], gdec)
        hi = hs(rt[ix:])
        low = hs(rt[1:ix])
        # Get response and stop accuracy information
        gacc = np.where(rt < self.tb, 0, 1)
        return [gacc, hi[~np.isnan(hi)], low[~np.isnan(low)]]


    def analyze_re_data(self, dv, p):
        import pandas as pd
        prob = self.percentiles
        ssd = self.ssd
        tb = self.tb
        nl = self.nlevels
        nss = self.nss
        nssd = self.nssd
        ntot = self.ntot
        nsstot = nss * nssd
        # get condition labels
        conditions = np.sort(self.fitparams['labels'] * int(ntot))
        # assign go trials pseudo ssd of 1000
        delays = np.append(np.array([[c] * nss for c in sd]), [1000] * nss * nssd)
        ttype = np.tile(array([['stop'] * nsstot + ['go'] * nsstot]), nl)

        DVg, DVs = dv
        if 'sso' in list(p):
            ssd = ssd + p['sso']
        gdec = self.go_resp(DVg, p['a'])
        sdec = self.ss_resp(DVs, p['a'])

        gort = self.RT(p['tr'], gdec)
        ssrt = self.RT(ssd, sdec)
        ert = np.tile(gort[:, :nss], nssd).reshape(nl, nssd, nss)

        # compare go proc. rt to ssrt and get error resp. (go proc win err)
        ssgo = map((lambda z: z[0][z[0] < z[1]]), zip(ert, ssrt))
        gresp = np.where(gort[:, nss * nssd:] < tb, 1, 0)
        ssresp = np.where(ert < ssrt, 1, 0).reshape(nl, nsstot)

        resp = np.concatenate([ssresp, gresp], axis=1)
        rt = np.concatenate([ert.reshape(nl, nsstot), gort[:, :nsstot]], axis=1)

        out = {'ssd': np.tile(delays, nl),
               'response': np.hstack(resp),
               'rt': np.hstack(rt),
               'cond': conditions,
               'ttype': ttype}

        return pd.DataFrame(out)



#!/usr/local/bin/env python
from __future__ import division
import os
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(font="Helvetica")

def get_cpals(name='all'):
    rpal = lambda nc: sns.blend_palette(['#e88379', '#c0392b'], n_colors=nc)
    bpal = lambda nc: sns.blend_palette(['#81aedb', '#3A539B'], n_colors=nc)
    gpal = lambda nc: sns.blend_palette(['#65b88f', '#27ae60'], n_colors=nc)
    ppal = lambda nc: sns.blend_palette(['#848bb6', "#9B59B6"], n_colors=nc)
    heat = lambda nc: sns.blend_palette(['#f39c12', '#c0392b'], n_colors=nc)
    cool = lambda nc: sns.blend_palette(["#4168B7", "#27ae60"], n_colors=nc)
    slate = lambda nc: sns.blend_palette(['#95A5A6', "#6C7A89"], n_colors=nc)
    color_dict = {'rpal': rpal, 'bpal': bpal, 'gpal': gpal, 'ppal': ppal, 'heat': heat, 'cool': cool, 'slate': slate}
    if name == 'all':
        return color_dict
    else:
        return color_dict[name]


def style_params(context='notebook'):
    colors = ["#1abc9c", "#16a085", "#2ecc71", "#27ae60",
              "#3498db", "#2980b9", "#4168B7", "#3A539B",
              "#9B59B6", "#8E44AD", "#663399", "#674172",
              "ghostwhite", '#95A5A6', '#6C7A89', "#34495e", "#2c3e50",
              '#E26A6A', "#e16256", "#e74c3c", "#ca4440", "#c0392b", "#bd4a4c",
              "#f1c40f", "#f39c12", "#e67e22"]

    return {'colors': colors, 'greens': colors[:4], 'blues': colors[4:8], 'purples': colors[8:12], 'grays': colors[12:17], 'reds': colors[17:22], 'yellows': colors[22:]}


def get_cmaps():
    block = ['Blues', 'BuGn', 'BuPu', 'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd', 'PuBu',
             'PuBuGn', 'PuRd', 'Purples', 'RdPu', 'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd'],
    seq = ['afmhot', 'autumn', 'bone', 'cool', 'copper',
           'gist_heat', 'gray', 'hot', 'pink', 'summer', 'winter']
    div = ['BrBG', 'bwr', 'coolwarm', 'PiYG', 'PRGn', 'PuOr',
           'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn', 'Spectral', 'seismic']
    qual = ['Accent', 'Dark2', 'Paired', 'Pastel1',
            'Pastel2', 'Set1', 'Set2', 'Set3']

    misc = ['gist_earth', 'terrain', 'ocean', 'gist_stern', 'brg', 'CMRmap', 'cubehelix', 'gnuplot',
            'gnuplot2', 'gist_ncar', 'nipy_spectral', 'jet', 'rainbow', 'gist_rainbow', 'hsv', 'flag', 'prism']
    cmaps = {'block': block, 'div': div,
             'seq': seq, 'qual': qual, 'misc': misc}
    return cmaps



#!/usr/bin/env python
from numpy.random import randint
from lmfit import report_fit, fit_report
from copy import deepcopy
import os


def get_one():

    msgs = ["Optimize On, Wayne",
            "Optimize On, Garth",
            "See, it's not that simplex...",
            "I wish you a merry fit, and a happy Nature paper",
            "It'll probably work this time",
            "'They dont think it be like it is, but it do' -Oscar Gamble",
            "Check your zoomfile before optimizing. It should be in your computer",
            "It's IN the computer?",
            "What is this... a model for ANTS!?"]

    return msgs[randint(0, len(msgs))]


def saygo(depends_on={}, labels=[], kind='radd', fit_on='subjects', dynamic='hyp'):

    pdeps = depends_on.keys()
    deplist = []
    pdep = describe_model(depends_on)

    if 'x' in kind:
        bias = '(w/ %s dynamic bias)' % dynamic
    else:
        bias = ""
    dep = depends_on.values()[0]
    lbls = ', '.join(labels)
    msg = get_one()
    strings = (kind, bias, fit_on, pdep, dep, lbls, msg)

    print("""
      Model is prepared to fit %s model %s to %s data,
      allowing %s to vary across levels of %s (%s)

      %s \n""" % strings)

    return True


def basin_accept_fun(x, f, accepted):
    print("at minimum %.4f accepted %d" % (f, int(accepted)))


def describe_model(depends_on=None):

    pdeps = depends_on.keys()
    deplist = []
    if 'a' in pdeps:
        deplist.append('boundary')
    if 'tr' in pdeps:
        deplist.append('onset')
    if 'v' in pdeps:
        deplist.append('drift')
    if 'xb' in pdeps:
        deplist.append('xbias')

    if len(pdeps) > 1:
        pdep = ' and '.join(deplist)
    else:
        pdep = deplist[0]

    return pdep


def global_logger(logs):

    arr_str = lambda x: ', '.join(str(elem)[:6] for elem in x)
    str_str = lambda x: ', '.join(str(elem) for elem in x)
    single_str = lambda x: ', '.join(str(elem) for elem in x)
    # create a dictionary string that can be copied and pasted into cell
    #pdict=[':'.join(["'"+str(k)+"'",str(v)]) for k,v in logs['popt'].items()]
    popt = ''.join(['popt =', repr(logs['popt'])])
    fmin_str = 'fmin = %s' % str(logs['fmin'])[:6]
    yhat_str = 'yhat = array([' + arr_str(logs['yhat']) + '])'
    y_str = 'y = array([' + arr_str(logs['y']) + '])'
    cost_str = 'cost = %s' % str(logs['cost'])[:6]

    with open('global_min_report.txt', 'a') as f:
        f.write('==' * 20 + '\n')
        f.write(popt + '\n')
        f.write(fmin_str + '\n')
        f.write(cost_str + '\n')
        f.write(yhat_str + '\n')
        f.write(y_str + '\n')
        f.write('==' * 20 + '\n')


def logger(param_report, finfo={}, depends_on={}, pdict={}, is_flat=True, log_arrays={}, kind=None, dynamic=None, fit_on=None, pc_map=None):

    wts, y, yhat = log_arrays['wts'], log_arrays['y'], log_arrays['yhat']
    if is_flat:
        fit_on = ' '.join([fit_on, 'FLAT'])
    else:
        fit_on = ' '.join([fit_on, 'FULL'])
    pkeys = depends_on.keys()
    pvals = depends_on.values()
    fname = '_'.join(['./' + kind, '_'.join(pkeys) + '.txt'])
    model_id = "MODEL: %s" % kind
    if 'x' in kind:
        model_id = ' ('.join([model_id, dynamic]) + ')'
    dep_id = "%s DEPENDS ON %s" % (pvals[0], str(tuple(pkeys)))
    wts_str = 'wts = array([' + ', '.join(str(elem)[:6] for elem in wts) + '])'
    yhat_str = 'yhat = array([' + ', '.join(str(elem)[:6] for elem in yhat) + '])'
    y_str = 'y = array([' + ', '.join(str(elem)[:6] for elem in y) + '])'
    with open(fname, 'a') as f:
        f.write('\n\n')
        f.write('==' * 30 + '\n\n')
        f.write(str(fit_on) + '\n')
        f.write(str(model_id) + '\n')
        f.write(str(dep_id) + '\n\n')
        f.write(wts_str + '\n\n')
        f.write(yhat_str + '\n\n')
        f.write(y_str + '\n\n')
        f.write('--' * 30 + '\n')
        f.write("FIT REPORT")
        f.write('\n' + '--' * 30 + '\n')
        f.write(param_report)
        f.write('\n' + '--' * 30 + '\n')
        f.write('='.join(['popt', repr(pdict)]) + '\n')
        f.write('AIC: %.8f' % finfo['AIC'] + '\n')
        f.write('BIC: %.8f' % finfo['BIC'] + '\n')
        f.write('chi: %.8f' % finfo['chi'] + '\n')
        f.write('rchi: %.8f' % finfo['rchi'] + '\n')
        f.write('converged: %s' % finfo['cnvrg'] + '\n\n')
        f.write('==' * 30 + '\n\n\n')
    return finfo



#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from radd import vis
from copy import deepcopy
import matplotlib as mpl
from IPython.display import display, Latex


def plot_traces_rts(p, all_traces, rts, names=['A', 'B', 'C', 'D'], tb=1000):

    rtkeys = np.sort(rts.keys())
    rt_dists = [np.asarray(rts[k])*1e3-(np.mean(p['tr'])*1e3) for k in rtkeys]
    tb = np.ceil(np.max([np.max(rti) if len(rti)>0 else 0 for rti in rt_dists]))+100

    sns.set(style='white', font_scale=1.2)
    f, axes = vis.build_multi_axis(p, tb=tb)
    clrs = sns.color_palette('muted', 5)

    for i in xrange(len(all_traces)):
        for ii, ax in enumerate(axes.flatten()):
            x=np.arange(len(all_traces[i][ii]))
            ax.plot(x, all_traces[i][ii], color=clrs[ii], alpha=.51, lw=.75)

    for i, ax in enumerate(axes.flatten()):
        divider = make_axes_locatable(ax)
        axx = divider.append_axes("top", size=.7, pad=0.01, sharex=ax)
        for spine in ['top', 'left', 'bottom', 'right']:
            axx.spines[spine].set_visible(False)
        axx.set_xticklabels([])
        axx.set_yticklabels([])
        if len(rt_dists[i])<=1:
            continue
        sns.distplot(rt_dists[i], ax=axx, label=k, color=clrs[i])
        text_str='$\mu_{%s}=%.fms$'%(names[i], np.mean(rt_dists[i]))
        ax.text(x[0]-50, np.mean(p['a'])-.06, text_str, fontsize=15)


def plot_summary(outcomes, titles=['Order of Choices','Number of Choices per Card', 'Change in Q(card)',
    'Change in P(card)', '$v^G_t$', '$v^N_t$'], plot_traces=False, p=None, tb=1000):

    sns.set_palette('muted')
    f, axes = plt.subplots(3, 2, figsize=(14,16))
    a1, a2, a3, a4, a5, a6 = axes.flatten()
    choices, rts, all_traces, qdict, choicep, vdhist, vihist = outcomes

    names = np.sort(qdict.keys())
    name_labels = [name.upper() for name in names]

    a1.plot(choices, lw=0, marker='o')
    a1.set_ylim(-.5, 3.5); a1.set_yticks(np.arange(4))
    a1.set_yticklabels(name_labels)

    a2.hist(np.asarray(choices))
    a2.set_xticks(np.arange(4))
    a2.set_xticklabels(name_labels)

    for i, n in enumerate(names):
        a3.plot(np.array(qdict[n])*100, label=name_labels[i])
        a4.plot(choicep[n], label=name_labels[i])
        a5.plot(vdhist[n], label=name_labels[i])
        a6.plot(vihist[n], label=name_labels[i])

    a3.legend(loc=0)
    a4.legend(loc=0)
    a5.legend(loc=0)
    a6.legend(loc=0)

    f.subplots_adjust(hspace=.35, wspace=.4)

    for i, ax in enumerate(axes.flatten()):
        ax.set_title(titles[i])
        sns.despine(ax=ax)
    if plot_traces and p is not None:
        plot_traces_rts(p, all_traces, rts, tb=tb)


def get_avg_slope_trace(traces, nalt=4):

    slope_func = lambda x0,x1,y0,y1: (y1-y0)/(x1-x0)
    rise, run, slopes=[], [], []
    for alt_n in range(nalt):
        y0 = 0
        y1 = np.mean([traces_i[alt_n][-1] for traces_i in traces])
        x0 = 0
        x1 = np.ceil(np.mean([traces_i[alt_n].shape[0] for traces_i in traces]))
        rise.append(y1)
        run.append(x1)
        slopes.append(slope_func(x0,x1,y0,y1))

    return rise, run, slopes


def gen_mappable(vals_to_map, cm='rainbow'):

    ncolors = len(vals_to_map)
    clrs = sns.color_palette(cm, ncolors)
    mycmap = mpl.colors.ListedColormap(clrs)

    vmin = np.min(vals_to_map)
    vmax = np.max(vals_to_map)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    scalar_mappable = mpl.cm.ScalarMappable(cmap=mycmap, norm=norm)
    scalar_mappable.set_array([])

    return scalar_mappable, vmin, vmax


def plot_reactivity_strategy(trialsdf, igtdf, cm='rainbow', save=False, pq='P'):

    if pq=='P':
        measure = 'Payoff ("P")'
    else:
        measure = 'Sensitivity ("Q")'

    n = trialsdf.agroup.unique().size
    a_go = np.sort(trialsdf.a_go.unique())
    sm, vmin, vmax = gen_mappable(vals_to_map=a_go, cm=cm)

    f, axes = plt.subplots(2,2, figsize=(14, 10))
    ax1, ax2, ax3, ax4 = axes.flatten()

    for grp, grpdf in trialsdf.groupby('agroup'):
        colr = sm.to_rgba(a_go[int(grp)-1])
        ax1.plot(grpdf.vdiff.values, color=colr)
        ax2.plot(grpdf.v_opt_diff.values, color=colr)
        sns.despine()

    divider = make_axes_locatable(ax1)
    cax = divider.append_axes("right", size="5%", pad=0.2)
    cb = plt.colorbar(sm, cax)
    sm.colorbar.set_ticks([vmin, vmax])
    cax.set_yticklabels([vmin, vmax])

    pvals_by_group = igtdf.groupby('agroup').mean().loc[:, pq].values
    reactivity = np.array([grpdf.vdiff.mean() for grp, grpdf in trialsdf.groupby('agroup')])
    strategy = np.array([grpdf.v_opt_diff.mean() for grp, grpdf in trialsdf.groupby('agroup')])

    for i in range(n):
        colr = sm.to_rgba(a_go[i])
        ax3.scatter(pvals_by_group[i], reactivity[i], color=colr, s=30)
        ax4.scatter(pvals_by_group[i], strategy[i], color=colr, s=30)

    ax1.set_title('$\Delta_{card} = V_{D}(t)-V_{I}(t)$', fontsize=19)
    ax2.set_title('$\Delta_{OS} = (\Delta_C + \Delta_D) - (\Delta_A + \Delta_B)$', fontsize=19)
    ax1.set_ylabel('Selected Channel Strength ($\Delta_{card}$)', fontsize=17)
    ax2.set_ylabel('Optimal-Subopt Ch. Strength ($\Delta_{OS}$)', fontsize=17)
    ax3.set_ylabel('$\mu \Delta_{card}$', fontsize=22)
    ax4.set_ylabel('$\mu \Delta_{OS}$', fontsize=22)

    plt.tight_layout()
    for ax in [ax1, ax2]:
        ax.set_xlabel('Trials', fontsize=22)
    for ax in [ax3, ax4]:
        ax.set_xlabel(measure, fontsize=22)
    f.subplots_adjust(wspace=.4, hspace=.3)
    if save:
        f.savefig('.'.join(['reactivity_strategy_', measure, 'png']), dpi=400)



#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
from radd.rl import visr, analyzer
from copy import deepcopy
from radd import theta

temporal_dynamics = lambda p, t: np.cosh(p['xb'][:, na] * t)
updateQ = lambda q, winner, r, A: q[winner][-1] + A*(r - q[winner][-1])
softmax_update = lambda q, name, B: np.exp(B*q[name][-1])/np.sum([np.exp(B*q[k][-1]) for k in q.keys()])

def run_full_sims(p, env=pd.DataFrame, alphas_go=[], alphas_no=None, betas=[], nblocks=2, nagents=100, si=.01, agent_list=[]):

    n_a = len(alphas_go)
    n_b = len(betas)
    trials = np.arange(1, (len(env)*nblocks)+1)

    if alphas_no is None:
        alphas_no=deepcopy(alphas_go)

    agroups = np.hstack([np.hstack([np.arange(n_a)]*n_b)]*nagents)
    bgroups = np.sort(np.hstack([np.sort(np.hstack([np.arange(n_b)]*n_a))]*nagents))
    agents = np.sort(np.hstack([np.arange(nagents)]*n_a*n_b))
    group = 0
    for agroup, bgroup, agent_i in zip(agroups, bgroups, agents):

        beta = betas[bgroup]
        a_go = alphas_go[agroup]
        a_no = alphas_no[agroup]

        pcopy=deepcopy(p)
        #p_rand = theta.random_inits(pcopy.keys())
        sim_out = run_trials(pcopy, env, nblocks=nblocks, si=si, a_go=a_go, a_no=a_no, beta=beta)
        choices, rts, all_traces, qdict, choicep, vd_all, vi_all = sim_out

        format_dict = {'agent':agent_i+1, 'trial':trials, 'a_go':a_go, 'a_no':a_no, 'adiff':a_go-a_no,
        'choices':choices, 'rts':rts, 'group': group, 'agroup': agroup+1, 'bgroup': bgroup+1,
        'qdict': qdict, 'choicep':choicep, 'vd_all':vd_all, 'vi_all':vi_all, 'beta':beta}

        format_dict_updated = analyzer.analyze_learning_dynamics(format_dict)
        igtdf, agdf = analyzer.format_dataframes(format_dict_updated)
        agent_list.append([agdf, igtdf])
        group += 1

    trial_df = pd.concat([ag[0] for ag in agent_list]) #.groupby(['group', 'trial']).mean().reset_index()
    igt_df = pd.concat([ag[1] for ag in agent_list], axis=1).T #.groupby('group').mean().reset_index()

    return [trial_df, igt_df]

def vectorize_params(p, pc_map, nresp=4):
    pvc = ['a', 'tr', 'vd', 'vi', 'xb']
    #remove conditional parameters from pvc
    [pvc.remove(param) for param in pc_map.keys()]
    for pkey in pvc:
        p[pkey]=p[pkey]*np.ones(nresp)
    for pkey, pkc in pc_map.items():
        if nresp==1:
            break
        elif pkc[0] not in p.keys():
            p[pkey] = p[pkey]*np.ones(len(pkc))
        else:
            p[pkey] = array([p[pc] for pc in pkc])
    return p


def rew_func(rprob):
    if rs()<rprob:
        return 1
    else:
        return 0


def run_trials(p, cards, nblocks=1, si=.01, a_go=.06, a_no=.06, beta=5):
    """simulate series of trials with learning
    Arguments:
        p (dict): parameter dictionary
        cards (DataFrame): pandas DF (ntrials x nalt) with choice outcome vaulues
    Returns:
        choices (list): choice made on each trial
        rts (dict): rt for each trial (winner rt)
        all_traces (list): execution process traces truncated to length of winner
        qdict (dict): sequence of Q-value updates for each alt
    """

    trials = cards.append([cards]*(nblocks-1)).reset_index()
    trials.rename(columns={'index':'t'}, inplace=True)
    ntrials=len(trials)
    choices, all_traces = [], []
    names = np.sort(cards.columns.values)
    rts={k:[] for k in names}
    qdict={k:[0] for k in names}
    choice_prob={k:[.25] for k in names}

    vdhist = pd.DataFrame(data=np.zeros((ntrials, len(names))), columns=names, index=np.arange(ntrials))
    vihist = vdhist.copy()

    for i in xrange(ntrials):
        vals = trials.iloc[i, 1:].values
        winner=np.nan
        iquit=0
        while np.isnan(winner) and iquit<20:
            execution = simulate_race(p, si=si)
            winner, rt, traces, p, qdict, choice_prob = analyze_multiresponse(execution, p, qdict=qdict, vals=vals, names=names, a_go=a_go, a_no=a_no, beta=beta, choice_prob=choice_prob)
            iquit+=1
            if np.isnan(np.mean(p['xb'])):
                iquit=30
        if winner>=len(names):
            winner = int(np.random.choice(np.arange(len(names))))
        vdhist.iloc[i, :] = p['vd']
        vihist.iloc[i, :] = p['vi']
        choice_name = names[winner]
        choices.append(winner); rts[choice_name].append(rt[winner]); all_traces.append(traces)

    return choices, rts, all_traces, qdict, choice_prob, vdhist, vihist


def simulate_race(p, pc_map={'vd': ['vd_a', 'vd_b', 'vd_c', 'vd_d'], 'vi': ['vi_a', 'vi_b', 'vi_c', 'vi_d']}, dt=.001, si=.01, tb=1.5, single_process=False, return_di=False):

    nresp = len(pc_map.values()[0])
    dx=np.sqrt(si*dt)
    p = vectorize_params(p, pc_map=pc_map, nresp=nresp)

    Tex = np.ceil((tb-p['tr'])/dt).astype(int)
    xtb = temporal_dynamics(p, np.cumsum([dt]*Tex.max()))

    if single_process:
        Pe = 0.5*(1 + (p['vd']-p['vi'])*dx/si)
        execution = xtb[0] * np.cumsum(np.where((rs((nresp, Tex.max())).T < Pe), dx, -dx).T, axis=1)

    else:
        Pd = 0.5*(1 + p['vd']*dx/si)
        Pi = 0.5*(1 + p['vi']*dx/si)
        direct = np.where((rs((nresp, Tex.max())).T < Pd),dx,-dx).T
        indirect = np.where((rs((nresp, Tex.max())).T < Pi),dx,-dx).T
        execution = xtb[0] * np.cumsum(direct-indirect, axis=1)
        if return_di:
            return np.cumsum(direct, axis=1), np.cumsum(indirect, axis=1), execution
    return execution


def analyze_multiresponse(execution, p, qdict={}, vals=[], names=[], a_go=.06, a_no=.06,  dt=.001, beta=5, choice_prob={}):
    """analyze multi-race execution processes"""

    nsteps_to_rt = np.argmax((execution.T>=p['a']).T, axis=1)
    rts = p['tr'] + nsteps_to_rt*dt

    # set non responses to 999
    rts[rts==p['tr'][0]]=999
    if np.all(rts==999):
        # if no response occurs, increase exponential bias (up to 3.0)
        if np.mean(p['xb']) <= 4.0:
            p['xb']=p['xb']*1.005
        return np.nan, rts, execution, p, qdict, choice_prob

    # get accumulator with fastest RT (winner) in each cond
    winner = np.argmin(rts)

    # get rt of winner in each cond
    winrt = rts[winner]

    # slice all traces at time the winner crossed boundary
    traces = [execution[i, :nsteps_to_rt[winner]] for i in xrange(len(rts))]

    reward = vals[winner]
    winner_name = names[winner]
    loser_names = names[names!=winner_name]

    # update action value
    qval = qdict[names[winner]][-1]
    if reward>=qval:
        alpha=a_go
    else:
        alpha=a_no

    Qt = updateQ(qdict, winner_name, reward, alpha)
    qdict[winner_name].append(Qt)

    for lname in loser_names:
        qdict[lname].append(qdict[lname][-1])

    #bound_expected = deepcopy(np.sum(p['vi']))
    for alt_i, name in enumerate(names):
        cp_old = choice_prob[name][-1]
        # update choice probability using boltzmann eq. w/ inv. temp beta
        cp_new = softmax_update(qdict, name, beta)
        choice_prob[name].append(cp_new)
        # calc. change in choice probability for alt_i
        delta_prob = cp_new - cp_old
        # update direct & indirect drift-rates with cp_delta
        p = reweight_drift(p, alt_i, delta_prob, a_go, a_no)
    #p['a'] = array([a_no*(bound_expected-np.sum(p['vi']))]*p['a'].size)

    return winner, rts, traces, p, qdict, choice_prob


def reweight_drift(p, alt_i, delta_prob, a_go, a_no):
    """ update direct & indirect drift-rates for multirace winner """

    vd_exp = p['vd'][alt_i]
    vi_exp = p['vi'][alt_i]

    p['vd'][alt_i] = vd_exp + (a_go*delta_prob)
    p['vi'][alt_i] = vi_exp + (a_no*-delta_prob)

    #p['vd'][alt_i] = vd_exp + (vd_exp*a_go)*delta_prob
    #p['vi'][alt_i] = vi_exp + (vi_exp*a_no)*-delta_prob

    return p



#!/usr/local/bin/env python
from __future__ import division
from future.utils import listvalues
import os
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from scipy.stats.mstats import mquantiles as mq
from radd.tools import theta, analyze, dfhandler


class RADDCore(object):

    """ Parent class for constructing attributes and methods used by
    of Model objects. Not meant to be used directly.

    Contains methods for building dataframes, generating observed data vectors
    that are entered into cost function during fitting as well as calculating
    summary measures and weight matrix for weighting residuals during optimization.

    TODO: COMPLETE DOCSTRINGS
    """

    def __init__(self, data=pd.DataFrame, kind='xdpm', inits=None, fit_on='average', depends_on={None: None}, dynamic='hyp', percentiles=np.array([.1, .3, .5, .7, .9]), weighted=True, verbose=False):

        self.verbose = verbose

        self.kind = kind
        self.fit_on = fit_on
        self.dynamic = dynamic
        self.weighted = weighted
        self.percentiles = percentiles
        self.tb = data[data.response == 1].rt.max()

        self.idx = list(data.idx.unique())
        self.nidx = len(self.idx)
        self.depends_on = depends_on

        if None in listvalues(depends_on):
            data['cond'] = 'flat'
            self.conds = ['cond']
            self.is_flat = True
        else:
            self.conds = listvalues(depends_on)
            self.is_flat = False

        self.nconds = len(self.conds)
        self.levels = [np.sort(data[cond].unique()) for cond in self.conds]
        self.nlevels = np.sum([len(lvls) for lvls in self.levels])
        self.groups = np.hstack([['idx'], self.conds]).tolist()

        # PARAMETER INITIALIZATION
        if inits is None:
            self.__get_default_inits__()
        else:
            self.inits = inits

        self.data = data
        # initialize dataframe handler
        self.handler = dfhandler.DataHandler(self)
        self.__prepare_fit__()
        

    def __prepare_fit__(self):
        """ model setup and initiates dataframes. Automatically run when Model object is initialized
            *   pc_map is a dict containing parameter names as keys with values
                    corresponding to the names given to that parameter in Parameters object
                    (see optmize.Optimizer).
            *   Parameters (p[pkey]=pval) that are constant across conditions are broadcast as [pval]*n.
                    Conditional parameters are treated as arrays with distinct values [V1, V2...Vn], one for
                    each condition.

            pc_map (dict):    keys: conditional parameter names (i.e. 'v')
                              values: keys + condition names ('v_bsl, v_pnl')

            |<--- PARAMETERS OBJECT [LMFIT] <-------- [IN]
            |
            |---> p = {'v_bsl': V1, 'v_pnl': V2...} --->|
                                                        |
            |<--- pc_map = {'v':['v_bsl', 'v_pnl']} <---|
            |
            |---> p['v'] = array([V1, V2]) -------> [OUT]
        """
        from radd.fit import Optimizer
        from radd.models import Simulator


        params = np.sort(list(self.inits)).tolist()
        cond_inits = lambda a, b: pd.Series(dict(zip(a, b)))
        self.pc_map = {}

        for cond_i in range(self.nconds):
            if self.is_flat:
                break
            for d in list(self.depends_on):
                params.remove(d)
                self.pc_map[d] = ['_'.join([d, l]) for l in self.levels[cond_i]]
                params.extend(self.pc_map[d])

        if 'ssd' in self.data.columns:
            self.__set_ssd__()

        # generate dataframes for observed data, popt, fitinfo, etc
        self.__make_dataframes__()

        # calculate costfx weights
        self.__set_wts__()

        # set fit parameters with default values
        self.set_fitparams()

        # set basinhopping parameters with default values
        self.set_basinparams()

        # initialize optimizer object for controlling fit routines
        # (updated with fitparams/basinparams whenever params are set)
        self.opt = Optimizer(fitparams=self.fitparams, basinparams=self.basinparams, kind=self.kind, inits=self.inits, depends_on=self.depends_on, pc_map=self.pc_map)

        # initialize model simulator, mainly accessed by the model optimizer object
        self.opt.simulator = Simulator(fitparams=self.fitparams, kind=self.kind, inits=self.inits, pc_map=self.pc_map)

        if self.verbose:
            self.is_prepared = messages.saygo(depends_on=self.depends_on, labels=self.levels,
            kind=self.kind, fit_on=self.fit_on, dynamic=self.dynamic)
        else:
            self.prepared = True


    def __make_dataframes__(self):
        """ wrapper for dfhandler.DataHandler.make_dataframes
        """
        self.handler.make_dataframes()
        # Group dataframe (nsubjects*nconds*nlevels x ndatapoints)
        self.observedDF = self.handler.observedDF.copy()
        # list (nsubjects long) of data arrays (nconds*nlevels x ndatapoints) to fit
        self.observed = self.handler.observed
        # list of flattened data arrays (averaged across conditions)
        self.observed_flat = self.handler.observed_flat

        self.datdf = self.handler.datdf
        self.dfvals = self.handler.dfvals

        # dataframe with same dim as observeddf for storing model predictions
        self.fits = self.handler.fits
        # dataframe with same dim as observeddf for storing fit info
        self.fitinfo = self.handler.fitinfo


    def __set_wts__(self, weight_presponse=True):
        """ wrapper for analyze functions used to calculate
        weights used in cost function
        """
        if self.weighted and self.fit_on == 'subjects':
            cost_wts, flat_cost_wts = analyze.get_subject_cost_weights(self.handler, weight_presponse=weight_presponse, percentiles=self.percentiles)
        elif self.weighted:
            cost_wts, flat_cost_wts = analyze.get_group_cost_weights(self.handler)
        else:
            self.flat_cost_wts = [np.ones_like(idat.flatten()) for idat in self.observed_flat]
            self.cost_wts = [np.ones_like(idat.flatten()) for idat in self.observed]
        # squeeze out any extra dimensions in wts arrays
        # ex: array([[1, 2, 3]]) --> array([1, 2, 3])
        self.cost_wts = [cw.squeeze() for cw in cost_wts]
        self.flat_cost_wts = [fcw.squeeze() for fcw in flat_cost_wts]


    def set_fitparams(self, get_params=False, **kwargs):
        """ dictionary of fit parameters, passed to Optimizer/Simulator objects
        """

        if not hasattr(self, 'fitparams'):
            # initialize with default values and first arrays in observed_flat, flat_cost_wts
            self.fitparams = {'ntrials': 10000, 'maxfev': 5000, 'maxiter': 500, 'disp': True, 'tol': 1.e-4,
                'method': 'nelder', 'niter': 500, 'tb': self.tb, 'percentiles': self.percentiles,
                'dynamic': self.dynamic, 'fit_on': self.fit_on, 'depends_on': self.depends_on}
            self.fitparams['idx'] = 0
            self.fitparams['y'] = self.observed_flat[0]
            self.fitparams['wts'] = self.flat_cost_wts[0]
            if hasattr(self, 'ssd'):
                self.fitparams['ssd'] = self.ssd[0]
                self.fitparams['nssd'] = self.fitparams['ssd'].size
        else:
            # fill with kwargs (i.e. y, wts, idx, etc) for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.fitparams[kw_arg] = kw_val
            if np.any([mk in self.kind for mk in ['dpm', 'irace', 'iact']]):
                self.fitparams['ssd'] = self.ssd[self.fitparams['idx']]
                self.fitparams['nssd'] = self.fitparams['ssd'].size
            self.opt.fitparams = self.fitparams
            self.opt.simulator.__update__(fitparams=self.opt.fitparams)

        if get_params:
            return self.fitparams


    def set_basinparams(self, get_params=False, **kwargs):
        """ dictionary of global fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'basinparams'):
            self.basinparams = {'nrand_inits': 30, 'nrand_samples': 5000, 'interval': 10, 'niter': 40, 'stepsize': .05, 'niter_success': 20, 'method': 'TNC', 'tol': 1.e-3, 'maxiter': 100, 'disp': True}
        else:
            # fill with kwargs for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.basinparams[kw_arg] = kw_val
            self.opt.basinparams = self.basinparams
        if get_params:
            return self.basinparams

    def __set_ssd__(self):
        """ set model attr "ssd" as list of np.arrays
        ssds to use when simulating data during optimization
        """
        grpdf = self.data.groupby(self.groups)
        get_stopdf = lambda df: df[df.ttype=='stop']
        get_df_ssds = lambda df: np.sort(get_stopdf(df).ssd.unique()*.001)
        self.ssd = [get_df_ssds(df) for _,df in grpdf]

    def __extract_popt_fitinfo__(self, finfo=None):
        """ wrapper for DataHandler.extract_popt_fitinfo()
        """
        popt = self.handler.extract_popt_fitinfo(finfo=finfo)
        return popt

    def rangl_data(self, data):
        """ wrapper for analyze.rangl_data
        """
        rangled = analyze.rangl_data(data, percentiles=self.percentiles, tb=self.tb)
        return rangled

    def resample_data(self, data):
        """ wrapper for analyze.resample_data
        """
        resampled = analyze.resample_data(data, n=100, data_style=self.data_style, tb=self.tb, kind=self.kind)
        return resampled

    def assess_fit(self, finfo=None):
        """ wrapper for analyze.assess_fit calculates and stores
        rchi, AIC, BIC and other fit statistics
        """
        return analyze.assess_fit(finfo)

    def params_io(self, p={}, io='w', iostr='popt'):
        """ read // write parameters dictionaries
        """
        if io == 'w':
            pd.Series(p).to_csv(''.join([iostr, '.csv']))
        elif io == 'r':
            ps = pd.read_csv(''.join([iostr, '.csv']), header=None)
            p = dict(zip(ps[0], ps[1]))
            return p

    def fits_io(self, fitparams, fits=[], io='w', iostr='fits'):
        """ read // write y, wts, yhat arrays
        """
        y = fitparams['y'].flatten()
        wts = fitparams['wts'].flatten()
        fits = fits.flatten()
        if io == 'w':
            index = np.arange(len(fits))
            df = pd.DataFrame({'y': y, 'wts': wts, 'yhat': fits}, index=index)
            df.to_csv(''.join([iostr, '.csv']))
        elif io == 'r':
            df = pd.read_csv(''.join([iostr, '.csv']), index_col=0)
            return df

    def __remove_outliers__(self, sd=1.5, verbose=False):
        self.data = analyze.remove_outliers(self.data, sd=sd, verbose=verbose)

    def __get_default_inits__(self):
        self.inits = theta.get_default_inits(kind=self.kind, dynamic=self.dynamic, depends_on=self.depends_on)

    def __get_optimized_params__(self, include_ss=False, fit_noise=False):
        params = theta.get_optimized_params(kind=self.kind, dynamic=self.dynamic, depends_on=self.depends_on)
        return params

    def __check_inits__(self, inits=None, pro_ss=False, fit_noise=False):
        if inits is None:
            inits = dict(deepcopy(self.inits))
        self.inits = theta.check_inits(inits=inits, pdep=list(self.depends_on), kind=self.kind, pro_ss=pro_ss, fit_noise=fit_noise)

    def __rename_bad_cols__(self):
        self.data = analyze.rename_bad_cols(self.data)



#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
from radd import vis
from copy import deepcopy


def igt_scores(choices):

    A=len(choices[choices==0])
    B=len(choices[choices==1])
    C=len(choices[choices==2])
    D=len(choices[choices==3])
    # payoff (P) score
    P = (C+D) - (A+B)
    # sensitivity (Q) score
    Q = (B+D) - (A+C)

    return P, Q


def analyze_learning_dynamics(fd):

    choices = fd['choices']
    vd_all = fd['vd_all']
    vi_all = fd['vi_all']
    vdiff_all = vd_all - vi_all
    qdict = fd['qdict']

    choice_vec = [np.sort(qdict.keys())[i] for i in choices]
    fd['choice']=choice_vec
    rts_copy = deepcopy(fd['rts'])
    fd['rt'] = [rts_copy[choice].pop(0) for i, choice in enumerate(choice_vec)]
    qcopy = deepcopy(qdict)
    fd['qval'] = [qcopy[choice].pop(0) for choice in choice_vec]

    fd['vd'] = [vd_all.loc[i, choice] for choice in choice_vec]
    fd['vi'] = [vi_all.loc[i, choice] for choice in choice_vec]
    fd['vdiff'] = [vdiff_all.loc[i, choice] for choice in choice_vec]

    vopt = vdiff_all['c'].values + vdiff_all['d'].values
    vsub = vdiff_all['a'].values + vdiff_all['b'].values

    vimp = vdiff_all['b'].values + vdiff_all['d'].values
    vnon = vdiff_all['a'].values + vdiff_all['c'].values
    fd['v_opt_diff'] = vopt - vsub
    fd['v_imp_diff'] = vimp - vnon

    #q_go = fd['qdict_go']
    #q_no = fd['qdict_no']
    #qgo_copy = deepcopy(q_go)
    #qno_copy = deepcopy(q_no)
    #fd['q_go'] = [qgo_copy[choice].pop(0) for choice in choice_vec]
    #fd['q_no'] = [qno_copy[choice].pop(0) for choice in choice_vec]

    return fd


def format_dataframes(fd):

    from collections import OrderedDict

    agdf_cols = ['agent', 'trial', 'agroup', 'bgroup', 'group', 'qval', 'vd', 'vi', 'vdiff',
                 'v_opt_diff', 'v_imp_diff', 'choice', 'rt', 'a_go', 'a_no', 'adiff', 'beta'] #'q_go', 'q_no',
    agdf = pd.DataFrame(OrderedDict((col, fd[col]) for col in agdf_cols))

    igtdf_cols=['agent', 'bgroup', 'agroup', 'group', 'a_go', 'a_no', 'beta', 'P', 'Q']
    fd['P'], fd['Q'] = igt_scores(np.asarray(fd['choices']))
    igtdf = pd.Series(OrderedDict((col, fd[col]) for col in igtdf_cols))

    return igtdf, agdf



from setuptools import setup
import numpy as np

setup(
    name='RADD',
    version='0.0.4',
    author='Kyle Dunovan, Timothy Verstynen',
    author_email='dunovank@gmail.com',
    url='http://github.com/CoAxLab/radd',
    packages=['radd', 'radd.tools', 'radd.rl'],
    description='RADD (Race Against Drift-Diffusion) is a python package for fitting & simulating cognitive models of reinforcement learning and decision-making',
    install_requires=['NumPy >=1.8.2', 'SciPy >= 0.16.1', 'matplotlib >= 1.4.3', 'seaborn>=0.5.1', 'pandas >= 0.12.0', 'lmfit>=0.9.0'],
    setup_requires=['NumPy >=1.8.2', 'SciPy >= 0.16.1', 'matplotlib >= 1.4.3', 'seaborn>=0.5.1', 'pandas >= 0.12.0', 'lmfit>=0.9.0'],
    include_package_data=True,
    include_dirs = [np.get_include()],
)
